@@ -7,17 +7,20 @@ use Encode;
 use HTML::Entities;
 use URI::Escape q{uri_escape_utf8};
 use POSIX;
+use Storable;
 use open qw{:utf8 :std};
 
 use vars qw{%config %links %oldlinks %pagemtime %pagectime %pagecase
-            %renderedfiles %oldrenderedfiles %pagesources %destsources
-	    %depends %hooks %forcerebuild $gettext_obj};
+	    %pagestate %renderedfiles %oldrenderedfiles %pagesources
+	    %destsources %depends %hooks %forcerebuild $gettext_obj};
 
 use Exporter q{import};
 our @EXPORT = qw(hook debug error template htmlpage add_depends pagespec_match
                  bestlink htmllink readfile writefile pagetype srcfile pagename
                  displaytime will_render gettext urlto targetpage
-                 %config %links %renderedfiles %pagesources %destsources);
+		 add_underlay
+                 %config %links %pagestate %renderedfiles
+                 %pagesources %destsources);
 our $VERSION = 2.00; # plugin interface version, next is ikiwiki version
 our $version='unknown'; # VERSION_AUTOREPLACE done by Makefile, DNE
 my $installdir=''; # INSTALLDIR_AUTOREPLACE done by Makefile, DNE
@@ -29,41 +32,30 @@ memoize("pagespec_translate");
 memoize("file_pruned");
 
 sub defaultconfig () { #{{{
-	wiki_file_prune_regexps => [qr/\.\./, qr/^\./, qr/\/\./,
+	return
+	wiki_file_prune_regexps => [qr/(^|\/)\.\.(\/|$)/, qr/^\./, qr/\/\./,
 		qr/\.x?html?$/, qr/\.ikiwiki-new$/,
 		qr/(^|\/).svn\//, qr/.arch-ids\//, qr/{arch}\//,
+		qr/(^|\/)_MTN\//,
 		qr/\.dpkg-tmp$/],
-	wiki_link_regexp => qr{
-		\[\[			# beginning of link
-		(?:
-			([^\]\|]+)	# 1: link text
-			\|		# followed by '|'
-		)?			# optional
-
-		([^\s\]#]+)		# 2: page to link to
-		(?:
-			\#		# '#', beginning of anchor
-			([^\s\]]+)	# 3: anchor text
-		)?			# optional
-
-		\]\]			# end of link
-	}x,
 	wiki_file_regexp => qr/(^[-[:alnum:]_.:\/+]+$)/,
 	web_commit_regexp => qr/^web commit (by (.*?(?=: |$))|from (\d+\.\d+\.\d+\.\d+)):?(.*)/,
 	verbose => 0,
 	syslog => 0,
 	wikiname => "wiki",
 	default_pageext => "mdwn",
+	htmlext => "html",
 	cgi => 0,
 	post_commit => 0,
 	rcs => '',
-	notify => 0,
 	url => '',
 	cgiurl => '',
 	historyurl => '',
 	diffurl => '',
 	rss => 0,
 	atom => 0,
+	allowrss => 0,
+	allowatom => 0,
 	discussion => 1,
 	rebuild => 0,
 	refresh => 0,
@@ -71,7 +63,6 @@ sub defaultconfig () { #{{{
 	w3mmode => 0,
 	wrapper => undef,
 	wrappermode => undef,
-	svnrepo => undef,
 	svnpath => "trunk",
 	gitorigin_branch => "origin",
 	gitmaster_branch => "master",
@@ -80,11 +71,13 @@ sub defaultconfig () { #{{{
 	pingurl => [],
 	templatedir => "$installdir/share/ikiwiki/templates",
 	underlaydir => "$installdir/share/ikiwiki/basewiki",
+	underlaydirs => [],
 	setup => undef,
 	adminuser => undef,
 	adminemail => undef,
-	plugin => [qw{mdwn inline htmlscrubber passwordauth openid signinedit
-	              lockedit conditional}],
+	plugin => [qw{mdwn link inline htmlscrubber passwordauth openid
+			signinedit lockedit conditional recentchanges}],
+	libdir => undef,
 	timeformat => '%c',
 	locale => undef,
 	sslcookie => 0,
@@ -93,8 +86,11 @@ sub defaultconfig () { #{{{
 	usedirs => 1,
 	numbacklinks => 10,
 	account_creation_password => "",
+	prefix_directives => 0,
+	hardlink => 0,
+	cgi_disable_uploads => 1,
 } #}}}
-   
+
 sub checkconfig () { #{{{
 	# locale stuff; avoid LC_ALL since it overrides everything
 	if (defined $ENV{LC_ALL}) {
@@ -108,6 +104,12 @@ sub checkconfig () { #{{{
 		}
 	}
 
+	if (ref $config{ENV} eq 'HASH') {
+		foreach my $val (keys %{$config{ENV}}) {
+			$ENV{$val}=$config{ENV}{$val};
+		}
+	}
+
 	if ($config{w3mmode}) {
 		eval q{use Cwd q{abs_path}};
 		error($@) if $@;
@@ -135,18 +137,30 @@ sub checkconfig () { #{{{
 		require IkiWiki::Rcs::Stub;
 	}
 
+	if (exists $config{umask}) {
+		umask(possibly_foolish_untaint($config{umask}));
+	}
+
 	run_hooks(checkconfig => sub { shift->() });
+
+	return 1;
 } #}}}
 
 sub loadplugins () { #{{{
+	if (defined $config{libdir}) {
+		unshift @INC, possibly_foolish_untaint($config{libdir});
+	}
+
 	loadplugin($_) foreach @{$config{plugin}};
-	
+
 	run_hooks(getopt => sub { shift->() });
 	if (grep /^-/, @ARGV) {
 		print STDERR "Unknown option: $_\n"
 			foreach grep /^-/, @ARGV;
 		usage();
 	}
+
+	return 1;
 } #}}}
 
 sub loadplugin ($) { #{{{
@@ -154,11 +168,21 @@ sub loadplugin ($) { #{{{
 
 	return if grep { $_ eq $plugin} @{$config{disable_plugins}};
 
+	foreach my $dir (defined $config{libdir} ? possibly_foolish_untaint($config{libdir}) : undef,
+	                 "$installdir/lib/ikiwiki") {
+		if (defined $dir && -x "$dir/plugins/$plugin") {
+			require IkiWiki::Plugin::external;
+			import IkiWiki::Plugin::external "$dir/plugins/$plugin";
+			return 1;
+		}
+	}
+
 	my $mod="IkiWiki::Plugin::".possibly_foolish_untaint($plugin);
 	eval qq{use $mod};
 	if ($@) {
 		error("Failed to load plugin $mod: $@");
 	}
+	return 1;
 } #}}}
 
 sub error ($;$) { #{{{
@@ -178,7 +202,7 @@ sub error ($;$) { #{{{
 
 sub debug ($) { #{{{
 	return unless $config{verbose};
-	log_message(debug => @_);
+	return log_message(debug => @_);
 } #}}}
 
 my $log_open=0;
@@ -187,20 +211,20 @@ sub log_message ($$) { #{{{
 
 	if ($config{syslog}) {
 		require Sys::Syslog;
-		unless ($log_open) {
+		if (! $log_open) {
 			Sys::Syslog::setlogsock('unix');
 			Sys::Syslog::openlog('ikiwiki', '', 'user');
 			$log_open=1;
 		}
-		eval {
+		return eval {
 			Sys::Syslog::syslog($type, "[$config{wikiname}] %s", join(" ", @_));
 		};
 	}
 	elsif (! $config{cgi}) {
-		print "@_\n";
+		return print "@_\n";
 	}
 	else {
-		print STDERR "@_\n";
+		return print STDERR "@_\n";
 	}
 } #}}}
 
@@ -230,7 +254,13 @@ sub pagetype ($) { #{{{
 	if ($page =~ /\.([^.]+)$/) {
 		return $1 if exists $hooks{htmlize}{$1};
 	}
-	return undef;
+	return;
+} #}}}
+
+sub isinternal ($) { #{{{
+	my $page=shift;
+	return exists $pagesources{$page} &&
+		$pagesources{$page} =~ /\._([^.]+)$/;
 } #}}}
 
 sub pagename ($) { #{{{
@@ -256,15 +286,36 @@ sub targetpage ($$) { #{{{
 sub htmlpage ($) { #{{{
 	my $page=shift;
 	
-	return targetpage($page, "html");
+	return targetpage($page, $config{htmlext});
 } #}}}
 
-sub srcfile ($) { #{{{
+sub srcfile_stat { #{{{
 	my $file=shift;
+	my $nothrow=shift;
 
-	return "$config{srcdir}/$file" if -e "$config{srcdir}/$file";
-	return "$config{underlaydir}/$file" if -e "$config{underlaydir}/$file";
-	error("internal error: $file cannot be found in $config{srcdir} or $config{underlaydir}");
+	return "$config{srcdir}/$file", stat(_) if -e "$config{srcdir}/$file";
+	foreach my $dir (@{$config{underlaydirs}}, $config{underlaydir}) {
+		return "$dir/$file", stat(_) if -e "$dir/$file";
+	}
+	error("internal error: $file cannot be found in $config{srcdir} or underlay") unless $nothrow;
+	return;
+} #}}}
+
+sub srcfile ($;$) { #{{{
+	return (srcfile_stat(@_))[0];
+} #}}}
+
+sub add_underlay ($) { #{{{
+	my $dir=shift;
+
+	if ($dir=~/^\//) {
+		unshift @{$config{underlaydirs}}, $dir;
+	}
+	else {
+		unshift @{$config{underlaydirs}}, "$config{underlaydir}/../$dir";
+	}
+
+	return 1;
 } #}}}
 
 sub readfile ($;$$) { #{{{
@@ -277,20 +328,17 @@ sub readfile ($;$$) { #{{{
 	}
 	
 	local $/=undef;
-	open (IN, $file) || error("failed to read $file: $!");
-	binmode(IN) if ($binary);
-	return \*IN if $wantfd;
-	my $ret=<IN>;
-	close IN || error("failed to read $file: $!");
+	open (my $in, "<", $file) || error("failed to read $file: $!");
+	binmode($in) if ($binary);
+	return \*$in if $wantfd;
+	my $ret=<$in>;
+	close $in || error("failed to read $file: $!");
 	return $ret;
 } #}}}
 
-sub writefile ($$$;$$) { #{{{
-	my $file=shift; # can include subdirs
-	my $destdir=shift; # directory to put file in
-	my $content=shift;
-	my $binary=shift;
-	my $writer=shift;
+sub prep_writefile ($$) {
+	my $file=shift;
+	my $destdir=shift;
 	
 	my $test=$file;
 	while (length $test) {
@@ -299,12 +347,8 @@ sub writefile ($$$;$$) { #{{{
 		}
 		$test=dirname($test);
 	}
-	my $newfile="$destdir/$file.ikiwiki-new";
-	if (-l $newfile) {
-		error("cannot write to a symlink ($newfile)");
-	}
 
-	my $dir=dirname($newfile);
+	my $dir=dirname("$destdir/$file");
 	if (! -d $dir) {
 		my $d="";
 		foreach my $s (split(m!/+!, $dir)) {
@@ -315,18 +359,37 @@ sub writefile ($$$;$$) { #{{{
 		}
 	}
 
+	return 1;
+}
+
+sub writefile ($$$;$$) { #{{{
+	my $file=shift; # can include subdirs
+	my $destdir=shift; # directory to put file in
+	my $content=shift;
+	my $binary=shift;
+	my $writer=shift;
+	
+	prep_writefile($file, $destdir);
+	
+	my $newfile="$destdir/$file.ikiwiki-new";
+	if (-l $newfile) {
+		error("cannot write to a symlink ($newfile)");
+	}
+	
 	my $cleanup = sub { unlink($newfile) };
-	open (OUT, ">$newfile") || error("failed to write $newfile: $!", $cleanup);
-	binmode(OUT) if ($binary);
+	open (my $out, '>', $newfile) || error("failed to write $newfile: $!", $cleanup);
+	binmode($out) if ($binary);
 	if ($writer) {
-		$writer->(\*OUT, $cleanup);
+		$writer->(\*$out, $cleanup);
 	}
 	else {
-		print OUT $content or error("failed writing to $newfile: $!", $cleanup);
+		print $out $content or error("failed writing to $newfile: $!", $cleanup);
 	}
-	close OUT || error("failed saving $newfile: $!", $cleanup);
+	close $out || error("failed saving $newfile: $!", $cleanup);
 	rename($newfile, "$destdir/$file") || 
 		error("failed renaming $newfile to $destdir/$file: $!", $cleanup);
+
+	return 1;
 } #}}}
 
 my %cleared;
@@ -352,6 +415,8 @@ sub will_render ($$;$) { #{{{
 		$cleared{$page}=1;
 	}
 	$destsources{$dest}=$page;
+
+	return 1;
 } #}}}
 
 sub bestlink ($$) { #{{{
@@ -363,6 +428,7 @@ sub bestlink ($$) { #{{{
 		# absolute links
 		$cwd="";
 	}
+	$link=~s/\/$//;
 
 	do {
 		my $l=$cwd;
@@ -394,7 +460,7 @@ sub bestlink ($$) { #{{{
 sub isinlinableimage ($) { #{{{
 	my $file=shift;
 	
-	$file=~/\.(png|gif|jpg|jpeg)$/i;
+	return $file =~ /\.(png|gif|jpg|jpeg)$/i;
 } #}}}
 
 sub pagetitle ($;$) { #{{{
@@ -454,20 +520,30 @@ sub abs2rel ($$) { #{{{
 	return $ret;
 } #}}}
 
-sub displaytime ($) { #{{{
+sub displaytime ($;$) { #{{{
 	my $time=shift;
+	my $format=shift;
+	if (! defined $format) {
+		$format=$config{timeformat};
+	}
 
 	# strftime doesn't know about encodings, so make sure
 	# its output is properly treated as utf8
-	return decode_utf8(POSIX::strftime(
-			$config{timeformat}, localtime($time)));
+	return decode_utf8(POSIX::strftime($format, localtime($time)));
 } #}}}
 
 sub beautify_url ($) { #{{{
 	my $url=shift;
 
-	$url =~ s!/index.html$!/!;
-	$url =~ s!^$!./!; # Browsers don't like empty links...
+	if ($config{usedirs}) {
+		$url =~ s!/index.$config{htmlext}$!/!;
+	}
+
+	# Ensure url is not an empty link, and
+	# if it's relative, make that explicit to avoid colon confusion.
+	if ($url !~ /\//) {
+		$url="./$url";
+	}
 
 	return $url;
 } #}}}
@@ -477,7 +553,7 @@ sub urlto ($$) { #{{{
 	my $from=shift;
 
 	if (! length $to) {
-		return beautify_url(baseurl($from));
+		return beautify_url(baseurl($from)."index.$config{htmlext}");
 	}
 
 	if (! $destsources{$to}) {
@@ -495,6 +571,8 @@ sub htmllink ($$$;@) { #{{{
 	my $link=shift;
 	my %opts=@_;
 
+	$link=~s/\/$//;
+
 	my $bestlink;
 	if (! $opts{forcesubpage}) {
 		$bestlink=bestlink($lpage, $link);
@@ -512,20 +590,21 @@ sub htmllink ($$$;@) { #{{{
 	}
 	
 	return "<span class=\"selflink\">$linktext</span>"
-		if length $bestlink && $page eq $bestlink;
+		if length $bestlink && $page eq $bestlink &&
+		   ! defined $opts{anchor};
 	
 	if (! $destsources{$bestlink}) {
 		$bestlink=htmlpage($bestlink);
 
 		if (! $destsources{$bestlink}) {
 			return $linktext unless length $config{cgiurl};
-			return "<span><a href=\"".
+			return "<span class=\"createlink\"><a href=\"".
 				cgiurl(
 					do => "create",
-					page => pagetitle(lc($link), 1),
+					page => lc($link),
 					from => $lpage
 				).
-				"\">?</a>$linktext</span>"
+				"\" rel=\"nofollow\">?</a>$linktext</span>"
 		}
 	}
 	
@@ -540,13 +619,41 @@ sub htmllink ($$$;@) { #{{{
 		$bestlink.="#".$opts{anchor};
 	}
 
-	return "<a href=\"$bestlink\">$linktext</a>";
+	my @attrs;
+	if (defined $opts{rel}) {
+		push @attrs, ' rel="'.$opts{rel}.'"';
+	}
+	if (defined $opts{class}) {
+		push @attrs, ' class="'.$opts{class}.'"';
+	}
+
+	return "<a href=\"$bestlink\"@attrs>$linktext</a>";
 } #}}}
 
-sub htmlize ($$$) { #{{{
+sub userlink ($) { #{{{
+	my $user=shift;
+
+	my $oiduser=eval { openiduser($user) };
+	if (defined $oiduser) {
+		return "<a href=\"$user\">$oiduser</a>";
+	}
+	else {
+		eval q{use CGI 'escapeHTML'};
+		error($@) if $@;
+
+		return htmllink("", "", escapeHTML(
+			length $config{userdir} ? $config{userdir}."/".$user : $user
+		), noimageinline => 1);
+	}
+} #}}}
+
+sub htmlize ($$$$) { #{{{
 	my $page=shift;
+	my $destpage=shift;
 	my $type=shift;
 	my $content=shift;
+	
+	my $oneline = $content !~ /\n/;
 
 	if (exists $hooks{htmlize}{$type}) {
 		$content=$hooks{htmlize}{$type}{call}->(
@@ -561,34 +668,39 @@ sub htmlize ($$$) { #{{{
 	run_hooks(sanitize => sub {
 		$content=shift->(
 			page => $page,
+			destpage => $destpage,
 			content => $content,
 		);
 	});
+	
+	if ($oneline) {
+		# hack to get rid of enclosing junk added by markdown
+		# and other htmlizers
+		$content=~s/^<p>//i;
+		$content=~s/<\/p>$//i;
+		chomp $content;
+	}
 
 	return $content;
 } #}}}
 
 sub linkify ($$$) { #{{{
-	my $lpage=shift; # the page containing the links
-	my $page=shift; # the page the link will end up on (different for inline)
+	my $page=shift;
+	my $destpage=shift;
 	my $content=shift;
 
-	$content =~ s{(\\?)$config{wiki_link_regexp}}{
-		defined $2
-			? ( $1 
-				? "[[$2|$3".($4 ? "#$4" : "")."]]" 
-				: htmllink($lpage, $page, linkpage($3),
-					anchor => $4, linktext => pagetitle($2)))
-			: ( $1 
-				? "[[$3".($4 ? "#$4" : "")."]]"
-				: htmllink($lpage, $page, linkpage($3),
-					anchor => $4))
-	}eg;
+	run_hooks(linkify => sub {
+		$content=shift->(
+			page => $page,
+			destpage => $destpage,
+			content => $content,
+		);
+	});
 	
 	return $content;
 } #}}}
 
-my %preprocessing;
+our %preprocessing;
 our $preprocess_preview=0;
 sub preprocess ($$$;$$) { #{{{
 	my $page=shift; # the page the data comes from
@@ -603,10 +715,11 @@ sub preprocess ($$$;$$) { #{{{
 
 	my $handle=sub {
 		my $escape=shift;
+		my $prefix=shift;
 		my $command=shift;
 		my $params=shift;
 		if (length $escape) {
-			return "[[$command $params]]";
+			return "[[$prefix$command $params]]";
 		}
 		elsif (exists $hooks{preprocess}{$command}) {
 			return "" if $scan && ! $hooks{preprocess}{$command}{scan};
@@ -614,7 +727,7 @@ sub preprocess ($$$;$$) { #{{{
 			# consider it significant.
 			my @params;
 			while ($params =~ m{
-				(?:(\w+)=)?		# 1: named parameter key?
+				(?:([-\w]+)=)?		# 1: named parameter key?
 				(?:
 					"""(.*?)"""	# 2: triple-quoted value
 				|
@@ -657,41 +770,81 @@ sub preprocess ($$$;$$) { #{{{
 					$command, $page, $preprocessing{$page}).
 				"]]";
 			}
-			my $ret=$hooks{preprocess}{$command}{call}->(
-				@params,
-				page => $page,
-				destpage => $destpage,
-				preview => $preprocess_preview,
-			);
+			my $ret;
+			if (! $scan) {
+				$ret=$hooks{preprocess}{$command}{call}->(
+					@params,
+					page => $page,
+					destpage => $destpage,
+					preview => $preprocess_preview,
+				);
+			}
+			else {
+				# use void context during scan pass
+				$hooks{preprocess}{$command}{call}->(
+					@params,
+					page => $page,
+					destpage => $destpage,
+					preview => $preprocess_preview,
+				);
+				$ret="";
+			}
 			$preprocessing{$page}--;
 			return $ret;
 		}
 		else {
-			return "\\[[$command $params]]";
+			return "[[$prefix$command $params]]";
 		}
 	};
 	
-	$content =~ s{
-		(\\?)		# 1: escape?
-		\[\[		# directive open
-		(\w+)		# 2: command
-		\s+
-		(		# 3: the parameters..
-			(?:
-				(?:\w+=)?		# named parameter key?
+	my $regex;
+	if ($config{prefix_directives}) {
+		$regex = qr{
+			(\\?)		# 1: escape?
+			\[\[(!)		# directive open; 2: prefix
+			([-\w]+)	# 3: command
+			(		# 4: the parameters..
+				\s+	# Must have space if parameters present
 				(?:
-					""".*?"""	# triple-quoted value
-					|
-					"[^"]+"		# single-quoted value
-					|
-					[^\s\]]+	# unquoted value
+					(?:[-\w]+=)?		# named parameter key?
+					(?:
+						""".*?"""	# triple-quoted value
+						|
+						"[^"]+"		# single-quoted value
+						|
+						[^\s\]]+	# unquoted value
+					)
+					\s*			# whitespace or end
+								# of directive
 				)
-				\s*			# whitespace or end
-							# of directive
-			)
-		*)		# 0 or more parameters
-		\]\]		# directive closed
-	}{$handle->($1, $2, $3)}sexg;
+			*)?		# 0 or more parameters
+			\]\]		# directive closed
+		}sx;
+	} else {
+		$regex = qr{
+			(\\?)		# 1: escape?
+			\[\[(!?)	# directive open; 2: optional prefix
+			([-\w]+)	# 3: command
+			\s+
+			(		# 4: the parameters..
+				(?:
+					(?:[-\w]+=)?		# named parameter key?
+					(?:
+						""".*?"""	# triple-quoted value
+						|
+						"[^"]+"		# single-quoted value
+						|
+						[^\s\]]+	# unquoted value
+					)
+					\s*			# whitespace or end
+								# of directive
+				)
+			*)		# 0 or more parameters
+			\]\]		# directive closed
+		}sx;
+	}
+
+	$content =~ s{$regex}{$handle->($1, $2, $3, $4)}eg;
 	return $content;
 } #}}}
 
@@ -712,6 +865,8 @@ sub indexlink () { #{{{
 	return "<a href=\"$config{url}\">$config{wikiname}</a>";
 } #}}}
 
+my $wikilock;
+
 sub lockwiki (;$) { #{{{
 	my $wait=@_ ? shift : 1;
 	# Take an exclusive lock on the wiki to prevent multiple concurrent
@@ -719,15 +874,15 @@ sub lockwiki (;$) { #{{{
 	if (! -d $config{wikistatedir}) {
 		mkdir($config{wikistatedir});
 	}
-	open(WIKILOCK, ">$config{wikistatedir}/lockfile") ||
+	open($wikilock, '>', "$config{wikistatedir}/lockfile") ||
 		error ("cannot write to $config{wikistatedir}/lockfile: $!");
-	if (! flock(WIKILOCK, 2 | 4)) { # LOCK_EX | LOCK_NB
+	if (! flock($wikilock, 2 | 4)) { # LOCK_EX | LOCK_NB
 		if ($wait) {
 			debug("wiki seems to be locked, waiting for lock");
 			my $wait=600; # arbitrary, but don't hang forever to 
 			              # prevent process pileup
 			for (1..$wait) {
-				return if flock(WIKILOCK, 2 | 4);
+				return if flock($wikilock, 2 | 4);
 				sleep 1;
 			}
 			error("wiki is locked; waited $wait seconds without lock being freed (possible stuck process or stale lock?)");
@@ -740,89 +895,134 @@ sub lockwiki (;$) { #{{{
 } #}}}
 
 sub unlockwiki () { #{{{
-	close WIKILOCK;
+	return close($wikilock) if $wikilock;
+	return;
 } #}}}
 
+my $commitlock;
+
 sub commit_hook_enabled () { #{{{
-	open(COMMITLOCK, "+>$config{wikistatedir}/commitlock") ||
-		error ("cannot write to $config{wikistatedir}/commitlock: $!");
-	if (! flock(COMMITLOCK, 1 | 4)) { # LOCK_SH | LOCK_NB to test
-		close COMMITLOCK;
+	open($commitlock, '+>', "$config{wikistatedir}/commitlock") ||
+		error("cannot write to $config{wikistatedir}/commitlock: $!");
+	if (! flock($commitlock, 1 | 4)) { # LOCK_SH | LOCK_NB to test
+		close($commitlock) || error("failed closing commitlock: $!");
 		return 0;
 	}
-	close COMMITLOCK;
+	close($commitlock) || error("failed closing commitlock: $!");
 	return 1;
 } #}}}
 
 sub disable_commit_hook () { #{{{
-	open(COMMITLOCK, ">$config{wikistatedir}/commitlock") ||
-		error ("cannot write to $config{wikistatedir}/commitlock: $!");
-	if (! flock(COMMITLOCK, 2)) { # LOCK_EX
+	open($commitlock, '>', "$config{wikistatedir}/commitlock") ||
+		error("cannot write to $config{wikistatedir}/commitlock: $!");
+	if (! flock($commitlock, 2)) { # LOCK_EX
 		error("failed to get commit lock");
 	}
+	return 1;
 } #}}}
 
 sub enable_commit_hook () { #{{{
-	close COMMITLOCK;
+	return close($commitlock) if $commitlock;
+	return;
 } #}}}
 
 sub loadindex () { #{{{
-	open (IN, "$config{wikistatedir}/index") || return;
-	while (<IN>) {
-		$_=possibly_foolish_untaint($_);
-		chomp;
-		my %items;
-		$items{link}=[];
-		$items{dest}=[];
-		foreach my $i (split(/ /, $_)) {
-			my ($item, $val)=split(/=/, $i, 2);
-			push @{$items{$item}}, decode_entities($val);
-		}
-
-		next unless exists $items{src}; # skip bad lines for now
-
-		my $page=pagename($items{src}[0]);
-		if (! $config{rebuild}) {
-			$pagesources{$page}=$items{src}[0];
-			$pagemtime{$page}=$items{mtime}[0];
-			$oldlinks{$page}=[@{$items{link}}];
-			$links{$page}=[@{$items{link}}];
-			$depends{$page}=$items{depends}[0] if exists $items{depends};
-			$destsources{$_}=$page foreach @{$items{dest}};
-			$renderedfiles{$page}=[@{$items{dest}}];
-			$pagecase{lc $page}=$page;
-		}
-		$oldrenderedfiles{$page}=[@{$items{dest}}];
-		$pagectime{$page}=$items{ctime}[0];
+	%oldrenderedfiles=%pagectime=();
+	if (! $config{rebuild}) {
+		%pagesources=%pagemtime=%oldlinks=%links=%depends=
+		%destsources=%renderedfiles=%pagecase=%pagestate=();
 	}
-	close IN;
+	my $in;
+	if (! open ($in, "<", "$config{wikistatedir}/indexdb")) {
+		if (-e "$config{wikistatedir}/index") {
+			system("ikiwiki-transition", "indexdb", $config{srcdir});
+			open ($in, "<", "$config{wikistatedir}/indexdb") || return;
+		}
+		else {
+			return;
+		}
+	}
+	my $ret=Storable::fd_retrieve($in);
+	if (! defined $ret) {
+		return 0;
+	}
+	my %index=%$ret;
+	foreach my $src (keys %index) {
+		my %d=%{$index{$src}};
+		my $page=pagename($src);
+		$pagectime{$page}=$d{ctime};
+		if (! $config{rebuild}) {
+			$pagesources{$page}=$src;
+			$pagemtime{$page}=$d{mtime};
+			$renderedfiles{$page}=$d{dest};
+			if (exists $d{links} && ref $d{links}) {
+				$links{$page}=$d{links};
+				$oldlinks{$page}=[@{$d{links}}];
+			}
+			if (exists $d{depends}) {
+				$depends{$page}=$d{depends};
+			}
+			if (exists $d{state}) {
+				$pagestate{$page}=$d{state};
+			}
+		}
+		$oldrenderedfiles{$page}=[@{$d{dest}}];
+	}
+	foreach my $page (keys %pagesources) {
+		$pagecase{lc $page}=$page;
+	}
+	foreach my $page (keys %renderedfiles) {
+		$destsources{$_}=$page foreach @{$renderedfiles{$page}};
+	}
+	return close($in);
 } #}}}
 
 sub saveindex () { #{{{
 	run_hooks(savestate => sub { shift->() });
 
+	my %hookids;
+	foreach my $type (keys %hooks) {
+		$hookids{$_}=1 foreach keys %{$hooks{$type}};
+	}
+	my @hookids=keys %hookids;
+
 	if (! -d $config{wikistatedir}) {
 		mkdir($config{wikistatedir});
 	}
-	my $newfile="$config{wikistatedir}/index.new";
+	my $newfile="$config{wikistatedir}/indexdb.new";
 	my $cleanup = sub { unlink($newfile) };
-	open (OUT, ">$newfile") || error("cannot write to $newfile: $!", $cleanup);
+	open (my $out, '>', $newfile) || error("cannot write to $newfile: $!", $cleanup);
+	my %index;
 	foreach my $page (keys %pagemtime) {
 		next unless $pagemtime{$page};
-		my $line="mtime=$pagemtime{$page} ".
-			"ctime=$pagectime{$page} ".
-			"src=$pagesources{$page}";
-		$line.=" dest=$_" foreach @{$renderedfiles{$page}};
-		my %count;
-		$line.=" link=$_" foreach grep { ++$count{$_} == 1 } @{$links{$page}};
+		my $src=$pagesources{$page};
+
+		$index{$src}={
+			ctime => $pagectime{$page},
+			mtime => $pagemtime{$page},
+			dest => $renderedfiles{$page},
+			links => $links{$page},
+		};
+
 		if (exists $depends{$page}) {
-			$line.=" depends=".encode_entities($depends{$page}, " \t\n");
+			$index{$src}{depends} = $depends{$page};
+		}
+
+		if (exists $pagestate{$page}) {
+			foreach my $id (@hookids) {
+				foreach my $key (keys %{$pagestate{$page}{$id}}) {
+					$index{$src}{state}{$id}{$key}=$pagestate{$page}{$id}{$key};
+				}
+			}
 		}
-		print OUT $line."\n" || error("failed writing to $newfile: $!", $cleanup);
 	}
-	close OUT || error("failed saving to $newfile: $!", $cleanup);
-	rename($newfile, "$config{wikistatedir}/index") ||
-		error("failed renaming $newfile to $config{wikistatedir}/index", $cleanup);
+	my $ret=Storable::nstore_fd(\%index, $out);
+	return if ! defined $ret || ! $ret;
+	close $out || error("failed saving to $newfile: $!", $cleanup);
+	rename($newfile, "$config{wikistatedir}/indexdb") ||
+		error("failed renaming $newfile to $config{wikistatedir}/indexdb", $cleanup);
+	
+	return 1;
 } #}}}
 
 sub template_file ($) { #{{{
@@ -831,7 +1031,7 @@ sub template_file ($) { #{{{
 	foreach my $dir ($config{templatedir}, "$installdir/share/ikiwiki/templates") {
 		return "$dir/$template" if -e "$dir/$template";
 	}
-	return undef;
+	return;
 } #}}}
 
 sub template_params (@) { #{{{
@@ -845,7 +1045,7 @@ sub template_params (@) { #{{{
 	my @ret=(
 		filter => sub {
 			my $text_ref = shift;
-			$$text_ref=&Encode::decode_utf8($$text_ref);
+			${$text_ref} = decode_utf8(${$text_ref});
 		},
 		filename => $filename,
 		loop_context_vars => 1,
@@ -857,7 +1057,7 @@ sub template_params (@) { #{{{
 
 sub template ($;@) { #{{{
 	require HTML::Template;
-	HTML::Template->new(template_params(@_));
+	return HTML::Template->new(template_params(@_));
 } #}}}
 
 sub misctemplate ($$;@) { #{{{
@@ -883,12 +1083,13 @@ sub hook (@) { # {{{
 	my %param=@_;
 	
 	if (! exists $param{type} || ! ref $param{call} || ! exists $param{id}) {
-		error "hook requires type, call, and id parameters";
+		error 'hook requires type, call, and id parameters';
 	}
 
 	return if $param{no_override} && exists $hooks{$param{type}}{$param{id}};
 	
 	$hooks{$param{type}}{$param{id}}=\%param;
+	return 1;
 } # }}}
 
 sub run_hooks ($$) { # {{{
@@ -910,6 +1111,8 @@ sub run_hooks ($$) { # {{{
 			$sub->($hooks{$type}{$id}{call});
 		}
 	}
+
+	return 1;
 } #}}}
 
 sub globlist_to_pagespec ($) { #{{{
@@ -925,9 +1128,9 @@ sub globlist_to_pagespec ($) { #{{{
 		}
 	}
 
-	my $spec=join(" or ", @spec);
+	my $spec=join(' or ', @spec);
 	if (@skip) {
-		my $skip=join(" and ", @skip);
+		my $skip=join(' and ', @skip);
 		if (length $spec) {
 			$spec="$skip and ($spec)";
 		}
@@ -940,7 +1143,7 @@ sub globlist_to_pagespec ($) { #{{{
 
 sub is_globlist ($) { #{{{
 	my $s=shift;
-	$s=~/[^\s]+\s+([^\s]+)/ && $1 ne "and" && $1 ne "or";
+	return ( $s =~ /[^\s]+\s+([^\s]+)/ && $1 ne "and" && $1 ne "or" );
 } #}}}
 
 sub safequote ($) { #{{{
@@ -953,27 +1156,33 @@ sub add_depends ($$) { #{{{
 	my $page=shift;
 	my $pagespec=shift;
 	
+	return unless pagespec_valid($pagespec);
+
 	if (! exists $depends{$page}) {
 		$depends{$page}=$pagespec;
 	}
 	else {
 		$depends{$page}=pagespec_merge($depends{$page}, $pagespec);
 	}
+
+	return 1;
 } # }}}
 
 sub file_pruned ($$) { #{{{
 	require File::Spec;
 	my $file=File::Spec->canonpath(shift);
 	my $base=File::Spec->canonpath(shift);
-	$file=~s#^\Q$base\E/*##;
+	$file =~ s#^\Q$base\E/+##;
 
 	my $regexp='('.join('|', @{$config{wiki_file_prune_regexps}}).')';
-	$file =~ m/$regexp/;
+	return $file =~ m/$regexp/ && $file ne $base;
 } #}}}
 
 sub gettext { #{{{
 	# Only use gettext in the rare cases it's needed.
-	if (exists $ENV{LANG} || exists $ENV{LC_ALL} || exists $ENV{LC_MESSAGES}) {
+	if ((exists $ENV{LANG} && length $ENV{LANG}) ||
+	    (exists $ENV{LC_ALL} && length $ENV{LC_ALL}) ||
+	    (exists $ENV{LC_MESSAGES} && length $ENV{LC_MESSAGES})) {
 		if (! $gettext_obj) {
 			$gettext_obj=eval q{
 				use Locale::gettext q{textdomain};
@@ -1010,8 +1219,6 @@ sub pagespec_merge ($$) { #{{{
 } #}}}
 
 sub pagespec_translate ($) { #{{{
-	# This assumes that $page is in scope in the function
-	# that evalulates the translated pagespec code.
 	my $spec=shift;
 
 	# Support for old-style GlobLists.
@@ -1030,36 +1237,41 @@ sub pagespec_translate ($) { #{{{
 		|
 			\)		# )
 		|
-			\w+\([^\)]+\)	# command(params)
+			\w+\([^\)]*\)	# command(params)
 		|
 			[^\s()]+	# any other text
 		)
 		\s*		# ignore whitespace
 	}igx) {
 		my $word=$1;
-		if (lc $word eq "and") {
-			$code.=" &&";
+		if (lc $word eq 'and') {
+			$code.=' &&';
 		}
-		elsif (lc $word eq "or") {
-			$code.=" ||";
+		elsif (lc $word eq 'or') {
+			$code.=' ||';
 		}
 		elsif ($word eq "(" || $word eq ")" || $word eq "!") {
-			$code.=" ".$word;
+			$code.=' '.$word;
 		}
 		elsif ($word =~ /^(\w+)\((.*)\)$/) {
 			if (exists $IkiWiki::PageSpec::{"match_$1"}) {
-				$code.="IkiWiki::PageSpec::match_$1(\$page, ".safequote($2).", \@params)";
+				$code.="IkiWiki::PageSpec::match_$1(\$page, ".safequote($2).", \@_)";
 			}
 			else {
-				$code.=" 0";
+				$code.=' 0';
 			}
 		}
 		else {
-			$code.=" IkiWiki::PageSpec::match_glob(\$page, ".safequote($word).", \@params)";
+			$code.=" IkiWiki::PageSpec::match_glob(\$page, ".safequote($word).", \@_)";
 		}
 	}
 
-	return $code;
+	if (! length $code) {
+		$code=0;
+	}
+
+	no warnings;
+	return eval 'sub { my $page=shift; '.$code.' }';
 } #}}}
 
 sub pagespec_match ($$;@) { #{{{
@@ -1069,12 +1281,26 @@ sub pagespec_match ($$;@) { #{{{
 
 	# Backwards compatability with old calling convention.
 	if (@params == 1) {
-		unshift @params, "location";
+		unshift @params, 'location';
 	}
 
-	my $ret=eval pagespec_translate($spec);
-	return IkiWiki::FailReason->new("syntax error") if $@;
-	return $ret;
+	my $sub=pagespec_translate($spec);
+	return IkiWiki::FailReason->new("syntax error in pagespec \"$spec\"") if $@;
+	return $sub->($page, @params);
+} #}}}
+
+sub pagespec_valid ($) { #{{{
+	my $spec=shift;
+
+	my $sub=pagespec_translate($spec);
+	return ! $@;
+} #}}}
+	
+sub glob2re ($) { #{{{
+	my $re=quotemeta(shift);
+	$re=~s/\\\*/.*/g;
+	$re=~s/\\\?/./g;
+	return $re;
 } #}}}
 
 package IkiWiki::FailReason;
@@ -1087,7 +1313,9 @@ use overload ( #{{{
 ); #}}}
 
 sub new { #{{{
-	bless \$_[1], $_[0];
+	my $class = shift;
+	my $value = shift;
+	return bless \$value, $class;
 } #}}}
 
 package IkiWiki::SuccessReason;
@@ -1100,7 +1328,9 @@ use overload ( #{{{
 ); #}}}
 
 sub new { #{{{
-	bless \$_[1], $_[0];
+	my $class = shift;
+	my $value = shift;
+	return bless \$value, $class;
 }; #}}}
 
 package IkiWiki::PageSpec;
@@ -1110,7 +1340,7 @@ sub match_glob ($$;@) { #{{{
 	my $glob=shift;
 	my %params=@_;
 	
-	my $from=exists $params{location} ? $params{location} : "";
+	my $from=exists $params{location} ? $params{location} : '';
 	
 	# relative matching
 	if ($glob =~ m!^\./!) {
@@ -1119,25 +1349,30 @@ sub match_glob ($$;@) { #{{{
 		$glob="$from/$glob" if length $from;
 	}
 
-	# turn glob into safe regexp
-	$glob=quotemeta($glob);
-	$glob=~s/\\\*/.*/g;
-	$glob=~s/\\\?/./g;
-
-	if ($page=~/^$glob$/i) {
-		return IkiWiki::SuccessReason->new("$glob matches $page");
+	my $regexp=IkiWiki::glob2re($glob);
+	if ($page=~/^$regexp$/i) {
+		if (! IkiWiki::isinternal($page) || $params{internal}) {
+			return IkiWiki::SuccessReason->new("$glob matches $page");
+		}
+		else {
+			return IkiWiki::FailReason->new("$glob matches $page, but the page is an internal page");
+		}
 	}
 	else {
 		return IkiWiki::FailReason->new("$glob does not match $page");
 	}
 } #}}}
 
+sub match_internal ($$;@) { #{{{
+	return match_glob($_[0], $_[1], @_, internal => 1)
+} #}}}
+
 sub match_link ($$;@) { #{{{
 	my $page=shift;
 	my $link=lc(shift);
 	my %params=@_;
 
-	my $from=exists $params{location} ? $params{location} : "";
+	my $from=exists $params{location} ? $params{location} : '';
 
 	# relative matching
 	if ($link =~ m!^\.! && defined $from) {
@@ -1146,16 +1381,16 @@ sub match_link ($$;@) { #{{{
 		$link="$from/$link" if length $from;
 	}
 
-	my $links = $IkiWiki::links{$page} or return undef;
-	return IkiWiki::FailReason->new("$page has no links") unless @$links;
+	my $links = $IkiWiki::links{$page};
+	return IkiWiki::FailReason->new("$page has no links") unless $links && @{$links};
 	my $bestlink = IkiWiki::bestlink($from, $link);
-	foreach my $p (@$links) {
+	foreach my $p (@{$links}) {
 		if (length $bestlink) {
 			return IkiWiki::SuccessReason->new("$page links to $link")
 				if $bestlink eq IkiWiki::bestlink($page, $p);
 		}
 		else {
-			return IkiWiki::SuccessReason->new("$page links to page matching $link")
+			return IkiWiki::SuccessReason->new("$page links to page $p matching $link")
 				if match_glob($p, $link, %params);
 		}
 	}
@@ -1163,7 +1398,7 @@ sub match_link ($$;@) { #{{{
 } #}}}
 
 sub match_backlink ($$;@) { #{{{
-	match_link($_[1], $_[0], @_);
+	return match_link($_[1], $_[0], @_);
 } #}}}
 
 sub match_created_before ($$;@) { #{{{
@@ -1172,10 +1407,10 @@ sub match_created_before ($$;@) { #{{{
 
 	if (exists $IkiWiki::pagectime{$testpage}) {
 		if ($IkiWiki::pagectime{$page} < $IkiWiki::pagectime{$testpage}) {
-			IkiWiki::SuccessReason->new("$page created before $testpage");
+			return IkiWiki::SuccessReason->new("$page created before $testpage");
 		}
 		else {
-			IkiWiki::FailReason->new("$page not created before $testpage");
+			return IkiWiki::FailReason->new("$page not created before $testpage");
 		}
 	}
 	else {
@@ -1189,10 +1424,10 @@ sub match_created_after ($$;@) { #{{{
 
 	if (exists $IkiWiki::pagectime{$testpage}) {
 		if ($IkiWiki::pagectime{$page} > $IkiWiki::pagectime{$testpage}) {
-			IkiWiki::SuccessReason->new("$page created after $testpage");
+			return IkiWiki::SuccessReason->new("$page created after $testpage");
 		}
 		else {
-			IkiWiki::FailReason->new("$page not created after $testpage");
+			return IkiWiki::FailReason->new("$page not created after $testpage");
 		}
 	}
 	else {
@@ -1202,42 +1437,28 @@ sub match_created_after ($$;@) { #{{{
 
 sub match_creation_day ($$;@) { #{{{
 	if ((gmtime($IkiWiki::pagectime{shift()}))[3] == shift) {
-		return IkiWiki::SuccessReason->new("creation_day matched");
+		return IkiWiki::SuccessReason->new('creation_day matched');
 	}
 	else {
-		return IkiWiki::FailReason->new("creation_day did not match");
+		return IkiWiki::FailReason->new('creation_day did not match');
 	}
 } #}}}
 
 sub match_creation_month ($$;@) { #{{{
 	if ((gmtime($IkiWiki::pagectime{shift()}))[4] + 1 == shift) {
-		return IkiWiki::SuccessReason->new("creation_month matched");
+		return IkiWiki::SuccessReason->new('creation_month matched');
 	}
 	else {
-		return IkiWiki::FailReason->new("creation_month did not match");
+		return IkiWiki::FailReason->new('creation_month did not match');
 	}
 } #}}}
 
 sub match_creation_year ($$;@) { #{{{
 	if ((gmtime($IkiWiki::pagectime{shift()}))[5] + 1900 == shift) {
-		return IkiWiki::SuccessReason->new("creation_year matched");
+		return IkiWiki::SuccessReason->new('creation_year matched');
 	}
 	else {
-		return IkiWiki::FailReason->new("creation_year did not match");
-	}
-} #}}}
-
-sub match_user ($$;@) { #{{{
-	shift;
-	my $user=shift;
-	my %params=@_;
-
-	return IkiWiki::FailReason->new("cannot match user") unless exists $params{user};
-	if ($user eq $params{user}) {
-		return IkiWiki::SuccessReason->new("user is $user")
-	}
-	else {
-		return IkiWiki::FailReason->new("user is not $user");
+		return IkiWiki::FailReason->new('creation_year did not match');
 	}
 } #}}}
 
