@@ -21,6 +21,23 @@ sub printheader ($) { #{{{
 
 } #}}}
 
+sub showform ($$$$;@) { #{{{
+	my $form=shift;
+	my $buttons=shift;
+	my $session=shift;
+	my $cgi=shift;
+
+	if (exists $hooks{formbuilder}) {
+		run_hooks(formbuilder => sub {
+			shift->(form => $form, cgi => $cgi, session => $session,
+				buttons => $buttons);
+		});
+	}
+
+	printheader($session);
+	print misctemplate($form->title, $form->render(submit => $buttons), @_);
+}
+
 sub redirect ($$) { #{{{
 	my $q=shift;
 	my $url=shift;
@@ -43,73 +60,43 @@ sub check_canedit ($$$;$) { #{{{
 	run_hooks(canedit => sub {
 		return if defined $canedit;
 		my $ret=shift->($page, $q, $session);
-		if (defined $ret && $ret eq "") {
-			$canedit=1;
-		}
-		elsif (defined $ret) {
-			$canedit=0;
-			error($ret) unless $nonfatal;
+		if (defined $ret) {
+			if ($ret eq "") {
+				$canedit=1;
+			}
+			elsif (ref $ret eq 'CODE') {
+				$ret->() unless $nonfatal;
+				$canedit=0;
+			}
+			elsif (defined $ret) {
+				error($ret) unless $nonfatal;
+				$canedit=0;
+			}
 		}
 	});
 	return $canedit;
 } #}}}
 
-sub decode_form_utf8 ($) { #{{{
-	my $form = shift;
-	foreach my $f ($form->field) {
-		next if Encode::is_utf8(scalar $form->field($f));
-		$form->field(name  => $f,
-			     value => decode_utf8($form->field($f)),
-			     force => 1,
-			    );
+sub decode_cgi_utf8 ($) { #{{{
+	# decode_form_utf8 method is needed for 5.10
+	if ($] < 5.01) {
+		my $cgi = shift;
+		foreach my $f ($cgi->param) {
+			$cgi->param($f, map { decode_utf8 $_ } $cgi->param($f));
+		}
 	}
 } #}}}
 
-sub cgi_recentchanges ($) { #{{{
-	my $q=shift;
-	
-	# Optimisation: building recentchanges means calculating lots of
-	# links. Memoizing htmllink speeds it up a lot (can't be memoized
-	# during page builds as the return values may change, but they
-	# won't here.)
-	eval q{use Memoize};
-	error($@) if $@;
-	memoize("htmllink");
-
-	eval q{use Time::Duration};
-	error($@) if $@;
-
-	my $changelog=[rcs_recentchanges(100)];
-	foreach my $change (@$changelog) {
-		$change->{when} = concise(ago($change->{when}));
-
-		$change->{user} = userlink($change->{user});
-
-		my $is_excess = exists $change->{pages}[10]; # limit pages to first 10
-		delete @{$change->{pages}}[10 .. @{$change->{pages}}] if $is_excess;
-		$change->{pages} = [
-			map {
-				$_->{link} = htmllink("", "", $_->{page},
-					noimageinline => 1,
-					linktext => pagetitle($_->{page}));
-				$_;
-			} @{$change->{pages}}
-		];
-		push @{$change->{pages}}, { link => '...' } if $is_excess;
+sub decode_form_utf8 ($) { #{{{
+	if ($] >= 5.01) {
+		my $form = shift;
+		foreach my $f ($form->field) {
+			$form->field(name  => $f,
+			             value => decode_utf8($form->field($f)),
+		                     force => 1,
+			);
+		}
 	}
-
-	my $template=template("recentchanges.tmpl"); 
-	$template->param(
-		title => "RecentChanges",
-		indexlink => indexlink(),
-		wikiname => $config{wikiname},
-		changelog => $changelog,
-		baseurl => baseurl(),
-	);
-	run_hooks(pagetemplate => sub {
-		shift->(page => "", destpage => "", template => $template);
-	});
-	print $q->header(-charset => 'utf-8'), $template->output;
 } #}}}
 
 # Check if the user is signed in. If not, redirect to the signin form and
@@ -120,9 +107,7 @@ sub needsignin ($$) { #{{{
 
 	if (! defined $session->param("name") ||
 	    ! userinfo_get($session->param("name"), "regdate")) {
-	    	if (! defined $session->param("postsignin")) {
-			$session->param(postsignin => $ENV{QUERY_STRING});
-		}
+		$session->param(postsignin => $ENV{QUERY_STRING});
 		cgi_signin($q, $session);
 		cgi_savesession($session);
 		exit;
@@ -133,12 +118,12 @@ sub cgi_signin ($$) { #{{{
 	my $q=shift;
 	my $session=shift;
 
+	decode_cgi_utf8($q);
 	eval q{use CGI::FormBuilder};
 	error($@) if $@;
 	my $form = CGI::FormBuilder->new(
 		title => "signin",
 		name => "signin",
-		header => 1,
 		charset => "utf-8",
 		method => 'POST',
 		required => 'NONE',
@@ -157,25 +142,18 @@ sub cgi_signin ($$) { #{{{
 	$form->field(name => "do", type => "hidden", value => "signin",
 		force => 1);
 	
+	decode_form_utf8($form);
 	run_hooks(formbuilder_setup => sub {
-		shift->(form => $form, cgi => $q, session => $session);
+		shift->(form => $form, cgi => $q, session => $session,
+		        buttons => $buttons);
 	});
-	
 	decode_form_utf8($form);
 
-	if (exists $hooks{formbuilder}) {
-		run_hooks(formbuilder => sub {
-			shift->(form => $form, cgi => $q, session => $session,
-				buttons => $buttons);
-		});
-	}
-	else {
-		if ($form->submitted) {
-			$form->validate;
-		}
-		printheader($session);
-		print misctemplate($form->title, $form->render(submit => $buttons));
+	if ($form->submitted) {
+		$form->validate;
 	}
+
+	showform($form, $buttons, $session, $q);
 } #}}}
 
 sub cgi_postsignin ($$) { #{{{
@@ -191,9 +169,7 @@ sub cgi_postsignin ($$) { #{{{
 		exit;
 	}
 	else {
-		# This can occur, for example, if a user went to the signin
-		# url via a bookmark.
-		redirect($q, $config{url});
+		error(gettext("login failed, perhaps you need to turn on cookies?"));
 	}
 } #}}}
 
@@ -202,6 +178,17 @@ sub cgi_prefs ($$) { #{{{
 	my $session=shift;
 
 	needsignin($q, $session);
+	decode_cgi_utf8($q);
+	
+	# The session id is stored on the form and checked to
+	# guard against CSRF.
+	my $sid=$q->param('sid');
+	if (! defined $sid) {
+		$q->delete_all;
+	}
+	elsif ($sid ne $session->id) {
+		error(gettext("Your login session has expired."));
+	}
 
 	eval q{use CGI::FormBuilder};
 	error($@) if $@;
@@ -227,16 +214,19 @@ sub cgi_prefs ($$) { #{{{
 		],
 	);
 	my $buttons=["Save Preferences", "Logout", "Cancel"];
-
-	run_hooks(formbuilder_setup => sub {
-		shift->(form => $form, cgi => $q, session => $session);
-	});
 	
-	$form->field(name => "do", type => "hidden");
+	decode_form_utf8($form);
+	run_hooks(formbuilder_setup => sub {
+		shift->(form => $form, cgi => $q, session => $session,
+		        buttons => $buttons);
+	});
+	decode_form_utf8($form);
+	
+	$form->field(name => "do", type => "hidden", value => "prefs",
+		force => 1);
+	$form->field(name => "sid", type => "hidden", value => $session->id,
+		force => 1);
 	$form->field(name => "email", size => 50, fieldset => "preferences");
-	$form->field(name => "subscriptions", size => 50,
-		fieldset => "preferences",
-		comment => "(".htmllink("", "", "PageSpec", noimageinline => 1).")");
 	$form->field(name => "banned_users", size => 50,
 		fieldset => "admin");
 	
@@ -248,16 +238,12 @@ sub cgi_prefs ($$) { #{{{
 	if (! $form->submitted) {
 		$form->field(name => "email", force => 1,
 			value => userinfo_get($user_name, "email"));
-		$form->field(name => "subscriptions", force => 1,
-			value => userinfo_get($user_name, "subscriptions"));
 		if (is_admin($user_name)) {
 			$form->field(name => "banned_users", force => 1,
 				value => join(" ", get_banned_users()));
 		}
 	}
 	
-	decode_form_utf8($form);
-	
 	if ($form->submitted eq 'Logout') {
 		$session->delete();
 		redirect($q, $config{url});
@@ -268,11 +254,9 @@ sub cgi_prefs ($$) { #{{{
 		return;
 	}
 	elsif ($form->submitted eq 'Save Preferences' && $form->validate) {
-		foreach my $field (qw(email subscriptions)) {
-			if (defined $form->field($field) && length $form->field($field)) {
-				userinfo_set($user_name, $field, $form->field($field)) ||
-					error("failed to set $field");
-			}
+		if (defined $form->field('email')) {
+			userinfo_set($user_name, 'email', $form->field('email')) ||
+				error("failed to set email");
 		}
 		if (is_admin($user_name)) {
 			set_banned_users(grep { ! is_admin($_) }
@@ -283,57 +267,49 @@ sub cgi_prefs ($$) { #{{{
 		$form->text(gettext("Preferences saved."));
 	}
 	
-	if (exists $hooks{formbuilder}) {
-		run_hooks(formbuilder => sub {
-			shift->(form => $form, cgi => $q, session => $session,
-				buttons => $buttons);
-		});
-	}
-	else {
-		printheader($session);
-		print misctemplate($form->title, $form->render(submit => $buttons));
-	}
+	showform($form, $buttons, $session, $q);
 } #}}}
 
 sub cgi_editpage ($$) { #{{{
 	my $q=shift;
 	my $session=shift;
+	
+	decode_cgi_utf8($q);
 
 	my @fields=qw(do rcsinfo subpage from page type editcontent comments);
 	my @buttons=("Save Page", "Preview", "Cancel");
-	
 	eval q{use CGI::FormBuilder};
 	error($@) if $@;
 	my $form = CGI::FormBuilder->new(
 		fields => \@fields,
-		header => 1,
 		charset => "utf-8",
 		method => 'POST',
-		validate => {
-			editcontent => '/.+/',
-		},
 		required => [qw{editcontent}],
 		javascript => 0,
 		params => $q,
 		action => $config{cgiurl},
+		header => 0,
 		table => 0,
 		template => scalar template_params("editpage.tmpl"),
 		wikiname => $config{wikiname},
 	);
 	
+	decode_form_utf8($form);
 	run_hooks(formbuilder_setup => sub {
-		shift->(form => $form, cgi => $q, session => $session);
+		shift->(form => $form, cgi => $q, session => $session,
+			buttons => \@buttons);
 	});
-	
 	decode_form_utf8($form);
 	
-	# This untaint is safe because titlepage removes any problematic
-	# characters.
-	my ($page)=$form->field('page');
-	$page=titlepage(possibly_foolish_untaint($page));
-	if (! defined $page || ! length $page || file_pruned($page, $config{srcdir}) || $page=~/^\//) {
+	# This untaint is safe because we check file_pruned.
+	my $page=$form->field('page');
+	$page=possibly_foolish_untaint($page);
+	if (! defined $page || ! length $page ||
+	    file_pruned($page, $config{srcdir}) || $page=~/^\//) {
 		error("bad page name");
 	}
+
+	my $baseurl=$config{url}."/".htmlpage($page);
 	
 	my $from;
 	if (defined $form->field('from')) {
@@ -345,20 +321,21 @@ sub cgi_editpage ($$) { #{{{
 	if (exists $pagesources{$page} && $form->field("do") ne "create") {
 		$file=$pagesources{$page};
 		$type=pagetype($file);
-		if (! defined $type) {
+		if (! defined $type || $type=~/^_/) {
 			error(sprintf(gettext("%s is not an editable page"), $page));
 		}
 		if (! $form->submitted) {
 			$form->field(name => "rcsinfo",
 				value => rcs_prepedit($file), force => 1);
 		}
+		$form->field(name => "editcontent", validate => '/.*/');
 	}
 	else {
 		$type=$form->param('type');
 		if (defined $type && length $type && $hooks{htmlize}{$type}) {
 			$type=possibly_foolish_untaint($type);
 		}
-		elsif (defined $from) {
+		elsif (defined $from && exists $pagesources{$from}) {
 			# favor the type of linking page
 			$type=pagetype($pagesources{$from});
 		}
@@ -367,13 +344,16 @@ sub cgi_editpage ($$) { #{{{
 		if (! $form->submitted) {
 			$form->field(name => "rcsinfo", value => "", force => 1);
 		}
+		$form->field(name => "editcontent", validate => '/.+/');
 	}
 
 	$form->field(name => "do", type => 'hidden');
+	$form->field(name => "sid", type => "hidden", value => $session->id,
+		force => 1);
 	$form->field(name => "from", type => 'hidden');
 	$form->field(name => "rcsinfo", type => 'hidden');
 	$form->field(name => "subpage", type => 'hidden');
-	$form->field(name => "page", value => pagetitle($page, 1), force => 1);
+	$form->field(name => "page", value => $page, force => 1);
 	$form->field(name => "type", value => $type, force => 1);
 	$form->field(name => "comments", type => "text", size => 80);
 	$form->field(name => "editcontent", type => "textarea", rows => 20,
@@ -381,8 +361,9 @@ sub cgi_editpage ($$) { #{{{
 	$form->tmpl_param("can_commit", $config{rcs});
 	$form->tmpl_param("indexlink", indexlink());
 	$form->tmpl_param("helponformattinglink",
-		htmllink("", "", "HelpOnFormatting", noimageinline => 1));
-	$form->tmpl_param("baseurl", baseurl());
+		htmllink($page, $page, "ikiwiki/formatting",
+			noimageinline => 1,
+			linktext => "FormattingHelp"));
 	
 	if ($form->submitted eq "Cancel") {
 		if ($form->field("do") eq "create" && defined $from) {
@@ -397,19 +378,46 @@ sub cgi_editpage ($$) { #{{{
 		return;
 	}
 	elsif ($form->submitted eq "Preview") {
-		$form->tmpl_param("page_preview",
-			htmlize($page, $type,
-			linkify($page, "",
+		my $new=not exists $pagesources{$page};
+		if ($new) {
+			# temporarily record its type
+			$pagesources{$page}=$page.".".$type;
+		}
+
+		my $content=$form->field('editcontent');
+
+		run_hooks(editcontent => sub {
+			$content=shift->(
+				content => $content,
+				page => $page,
+				cgi => $q,
+				session => $session,
+			);
+		});
+		my $preview=htmlize($page, $page, $type,
+			linkify($page, $page,
 			preprocess($page, $page,
-			filter($page, $page, $form->field('editcontent')), 0, 1))));
+			filter($page, $page, $content), 0, 1)));
+		run_hooks(format => sub {
+			$preview=shift->(
+				page => $page,
+				content => $preview,
+			);
+		});
+		$form->tmpl_param("page_preview", $preview);
+	
+		if ($new) {
+			delete $pagesources{$page};
+		}
+		# previewing may have created files on disk
+		saveindex();
 	}
-	else {
+	elsif ($form->submitted eq "Save Page") {
 		$form->tmpl_param("page_preview", "");
 	}
 	$form->tmpl_param("page_conflict", "");
 	
-	if (! $form->submitted || $form->submitted eq "Preview" || 
-	    ! $form->validate) {
+	if ($form->submitted ne "Save Page" || ! $form->validate) {
 		if ($form->field("do") eq "create") {
 			my @page_locs;
 			my $best_loc;
@@ -438,9 +446,10 @@ sub cgi_editpage ($$) { #{{{
 					$dir=~s![^/]+/+$!!;
 					push @page_locs, $dir.$page;
 				}
+			
+				push @page_locs, "$config{userdir}/$page"
+					if length $config{userdir};
 			}
-			push @page_locs, "$config{userdir}/$page"
-				if length $config{userdir};
 
 			@page_locs = grep {
 				! exists $pagecase{lc $_}
@@ -448,8 +457,16 @@ sub cgi_editpage ($$) { #{{{
 			if (! @page_locs) {
 				# hmm, someone else made the page in the
 				# meantime?
-				redirect($q, "$config{url}/".htmlpage($page));
-				return;
+				if ($form->submitted eq "Preview") {
+					# let them go ahead with the edit
+					# and resolve the conflict at save
+					# time
+					@page_locs=$page;
+				}
+				else {
+					redirect($q, "$config{url}/".htmlpage($page));
+					return;
+				}
 			}
 
 			my @editable_locs = grep {
@@ -462,16 +479,18 @@ sub cgi_editpage ($$) { #{{{
 			
 			my @page_types;
 			if (exists $hooks{htmlize}) {
-				@page_types=keys %{$hooks{htmlize}};
+				@page_types=grep { !/^_/ }
+					keys %{$hooks{htmlize}};
 			}
 			
 			$form->tmpl_param("page_select", 1);
 			$form->field(name => "page", type => 'select',
-				options => [ map { pagetitle($_, 1) } @editable_locs ],
-				value => pagetitle($best_loc, 1));
+				options => [ map { [ $_, pagetitle($_, 1) ] } @editable_locs ],
+				value => $best_loc);
 			$form->field(name => "type", type => 'select',
 				options => \@page_types);
 			$form->title(sprintf(gettext("creating %s"), pagetitle($page)));
+			
 		}
 		elsif ($form->field("do") eq "edit") {
 			check_canedit($page, $q, $session);
@@ -491,24 +510,36 @@ sub cgi_editpage ($$) { #{{{
 			$form->title(sprintf(gettext("editing %s"), pagetitle($page)));
 		}
 		
-		print $form->render(submit => \@buttons);
+		showform($form, \@buttons, $session, $q, forcebaseurl => $baseurl);
 	}
 	else {
 		# save page
 		check_canedit($page, $q, $session);
-		if (! -e "$config{srcdir}/$file" &&
-		    $form->field("do") ne "create") {
+	
+		# The session id is stored on the form and checked to
+		# guard against CSRF. But only if the user is logged in,
+		# as anonok can allow anonymous edits.
+		if (defined $session->param("name")) {
+			my $sid=$q->param('sid');
+			if (! defined $sid || $sid ne $session->id) {
+				error(gettext("Your login session has expired."));
+			}
+		}
+
+		my $exists=-e "$config{srcdir}/$file";
+
+		if ($form->field("do") ne "create" && ! $exists &&
+		    ! defined srcfile($file, 1)) {
 			$form->tmpl_param("page_gone", 1);
 			$form->field(name => "do", value => "create", force => 1);
 			$form->tmpl_param("page_select", 0);
 			$form->field(name => "page", type => 'hidden');
 			$form->field(name => "type", type => 'hidden');
 			$form->title(sprintf(gettext("editing %s"), $page));
-			print $form->render(submit => \@buttons);
+			showform($form, \@buttons, $session, $q, forcebaseurl => $baseurl);
 			return;
 		}
-		elsif (-e "$config{srcdir}/$file" &&
-		       $form->field("do") eq "create") {
+		elsif ($form->field("do") eq "create" && $exists) {
 			$form->tmpl_param("creation_conflict", 1);
 			$form->field(name => "do", value => "edit", force => 1);
 			$form->tmpl_param("page_select", 0);
@@ -519,14 +550,22 @@ sub cgi_editpage ($$) { #{{{
 				value => readfile("$config{srcdir}/$file").
 				         "\n\n\n".$form->field("editcontent"),
 				force => 1);
-			print $form->render(submit => \@buttons);
+			showform($form, \@buttons, $session, $q, forcebaseurl => $baseurl);
 			return;
 		}
 		
 		my $content=$form->field('editcontent');
-
+		run_hooks(editcontent => sub {
+			$content=shift->(
+				content => $content,
+				page => $page,
+				cgi => $q,
+				session => $session,
+			);
+		});
 		$content=~s/\r\n/\n/g;
 		$content=~s/\r/\n/g;
+		$content.="\n" if $content !~ /\n$/;
 
 		$config{cgi}=0; # avoid cgi error message
 		eval { writefile($file, $config{srcdir}, $content) };
@@ -541,7 +580,8 @@ sub cgi_editpage ($$) { #{{{
 			$form->field(name => "page", type => 'hidden');
 			$form->field(name => "type", type => 'hidden');
 			$form->title(sprintf(gettext("editing %s"), $page));
-			print $form->render(submit => \@buttons);
+			showform($form, \@buttons, $session, $q,
+				forcebaseurl => $baseurl);
 			return;
 		}
 		
@@ -553,13 +593,13 @@ sub cgi_editpage ($$) { #{{{
 				$message=$form->field('comments');
 			}
 			
-			if ($form->field("do") eq "create") {
+			if (! $exists) {
 				rcs_add($file);
 			}
 
 			# Prevent deadlock with post-commit hook by
 			# signaling to it that it should not try to
-			# do anything (except send commit mails).
+			# do anything.
 			disable_commit_hook();
 			$conflict=rcs_commit($file, $message,
 				$form->field("rcsinfo"),
@@ -585,7 +625,8 @@ sub cgi_editpage ($$) { #{{{
 			$form->field(name => "page", type => 'hidden');
 			$form->field(name => "type", type => 'hidden');
 			$form->title(sprintf(gettext("editing %s"), $page));
-			print $form->render(submit => \@buttons);
+			showform($form, \@buttons, $session, $q,
+				forcebaseurl => $baseurl);
 			return;
 		}
 		else {
@@ -600,6 +641,7 @@ sub cgi_getsession ($) { #{{{
 	my $q=shift;
 
 	eval q{use CGI::Session};
+	error($@) if $@;
 	CGI::Session->name("ikiwiki_session_".encode_utf8($config{wikiname}));
 	
 	my $oldmask=umask(077);
@@ -623,11 +665,14 @@ sub cgi (;$$) { #{{{
 	my $q=shift;
 	my $session=shift;
 
+	eval q{use CGI};
+	error($@) if $@;
+	$CGI::DISABLE_UPLOADS=$config{cgi_disable_uploads};
+
 	if (! $q) {
-		eval q{use CGI};
-		error($@) if $@;
-	
+		binmode(STDIN);
 		$q=CGI->new;
+		binmode(STDIN, ":utf8");
 	
 		run_hooks(cgi => sub { shift->($q) });
 	}
@@ -643,17 +688,9 @@ sub cgi (;$$) { #{{{
 		}
 	}
 	
-	# Things that do not need a session.
-	if ($do eq 'recentchanges') {
-		cgi_recentchanges($q);
-		return;
-	}
-	elsif ($do eq 'hyperestraier') {
-		cgi_hyperestraier();
-	}
-
 	# Need to lock the wiki before getting a session.
 	lockwiki();
+	loadindex();
 	
 	if (! $session) {
 		$session=cgi_getsession($q);
@@ -684,68 +721,25 @@ sub cgi (;$$) { #{{{
 		print gettext("You are banned.");
 		cgi_savesession($session);
 	}
-	elsif ($do eq 'signin') {
+
+	run_hooks(sessioncgi => sub { shift->($q, $session) });
+
+	if ($do eq 'signin') {
 		cgi_signin($q, $session);
 		cgi_savesession($session);
 	}
-	elsif (defined $session->param("postsignin")) {
-		cgi_postsignin($q, $session);
-	}
 	elsif ($do eq 'prefs') {
 		cgi_prefs($q, $session);
 	}
 	elsif ($do eq 'create' || $do eq 'edit') {
 		cgi_editpage($q, $session);
 	}
-	elsif ($do eq 'blog') {
-		my $page=decode_utf8($q->param('title'));
-		$page=~s/\///g; # no slashes in blog posts
-		# if the page already exists, munge it to be unique
-		my $from=$q->param('from');
-		my $add="";
-		while (exists $pagecase{lc($from."/".titlepage($page).$add)}) {
-			$add=1 unless length $add;
-			$add++;
-		}
-		$q->param('page', $page.$add);
-		# now run same as create
-		$q->param('do', 'create');
-		cgi_editpage($q, $session);
-	}
-	elsif ($do eq 'postsignin') {
-		error(gettext("login failed, perhaps you need to turn on cookies?"));
+	elsif (defined $session->param("postsignin") || $do eq 'postsignin') {
+		cgi_postsignin($q, $session);
 	}
 	else {
 		error("unknown do parameter");
 	}
 } #}}}
 
-sub userlink ($) { #{{{
-	my $user=shift;
-
-	eval q{use CGI 'escapeHTML'};
-	error($@) if $@;
-	if ($user =~ m!^https?://! &&
-	    eval q{use Net::OpenID::VerifiedIdentity; 1} && !$@) {
-		# Munge user-urls, as used by eg, OpenID.
-		my $oid=Net::OpenID::VerifiedIdentity->new(identity => $user);
-		my $display=$oid->display;
-		# Convert "user.somehost.com" to "user [somehost.com]".
-		if ($display !~ /\[/) {
-			$display=~s/^(.*?)\.([^.]+\.[a-z]+)$/$1 [$2]/;
-		}
-		# Convert "http://somehost.com/user" to "user [somehost.com]".
-		if ($display !~ /\[/) {
-			$display=~s/^https?:\/\/(.+)\/([^\/]+)$/$2 [$1]/;
-		}
-		$display=~s!^https?://!!; # make sure this is removed
-		return "<a href=\"$user\">".escapeHTML($display)."</a>";
-	}
-	else {
-		return htmllink("", "", escapeHTML(
-			length $config{userdir} ? $config{userdir}."/".$user : $user
-		), noimageinline => 1);
-	}
-} #}}}
-
 1
