@@ -4,21 +4,30 @@ package IkiWiki::Plugin::inline;
 
 use warnings;
 use strict;
+use Encode;
 use IkiWiki 2.00;
 use URI;
 
+my %knownfeeds;
+my %page_numfeeds;
+my @inline;
+my $nested=0;
+
 sub import { #{{{
 	hook(type => "getopt", id => "inline", call => \&getopt);
 	hook(type => "checkconfig", id => "inline", call => \&checkconfig);
+	hook(type => "sessioncgi", id => "inline", call => \&sessioncgi);
 	hook(type => "preprocess", id => "inline", 
 		call => \&IkiWiki::preprocess_inline);
 	hook(type => "pagetemplate", id => "inline",
 		call => \&IkiWiki::pagetemplate_inline);
+	hook(type => "format", id => "inline", call => \&format);
 	# Hook to change to do pinging since it's called late.
 	# This ensures each page only pings once and prevents slow
 	# pings interrupting page builds.
 	hook(type => "change", id => "inline", 
 		call => \&IkiWiki::pingurl);
+
 } # }}}
 
 sub getopt () { #{{{
@@ -28,6 +37,8 @@ sub getopt () { #{{{
 	GetOptions(
 		"rss!" => \$config{rss},
 		"atom!" => \$config{atom},
+		"allowrss!" => \$config{allowrss},
+		"allowatom!" => \$config{allowatom},
 	);
 }
 
@@ -43,6 +54,39 @@ sub checkconfig () { #{{{
 	}
 } #}}}
 
+sub format (@) { #{{{
+        my %params=@_;
+
+	# Fill in the inline content generated earlier. This is actually an
+	# optimisation.
+	$params{content}=~s{<div class="inline" id="([^"]+)"></div>}{
+		delete @inline[$1,]
+	}eg;
+	return $params{content};
+} #}}}
+
+sub sessioncgi () { #{{{
+	my $q=shift;
+	my $session=shift;
+
+	if ($q->param('do') eq 'blog') {
+		my $page=IkiWiki::titlepage(decode_utf8($q->param('title')));
+		$page=~s/(\/)/"__".ord($1)."__"/eg; # don't create subdirs
+		# if the page already exists, munge it to be unique
+		my $from=$q->param('from');
+		my $add="";
+		while (exists $IkiWiki::pagecase{lc($from."/".$page.$add)}) {
+			$add=1 unless length $add;
+			$add++;
+		}
+		$q->param('page', $page.$add);
+		# now go create the page
+		$q->param('do', 'create');
+		IkiWiki::cgi_editpage($q, $session);
+		exit;
+	}
+}
+
 # Back to ikiwiki namespace for the rest, this code is very much
 # internal to ikiwiki even though it's separated into a plugin.
 package IkiWiki;
@@ -59,22 +103,26 @@ sub preprocess_inline (@) { #{{{
 	my %params=@_;
 	
 	if (! exists $params{pages}) {
-		return "";
+		return "[[inline ".gettext("missing pages parameter")."]]";
 	}
 	my $raw=yesno($params{raw});
 	my $archive=yesno($params{archive});
-	my $rss=($config{rss} && exists $params{rss}) ? yesno($params{rss}) : $config{rss};
-	my $atom=($config{atom} && exists $params{atom}) ? yesno($params{atom}) : $config{atom};
+	my $rss=(($config{rss} || $config{allowrss}) && exists $params{rss}) ? yesno($params{rss}) : $config{rss};
+	my $atom=(($config{atom} || $config{allowatom}) && exists $params{atom}) ? yesno($params{atom}) : $config{atom};
 	my $quick=exists $params{quick} ? yesno($params{quick}) : 0;
 	my $feeds=exists $params{feeds} ? yesno($params{feeds}) : !$quick;
-	$feeds=0 if $params{preview};
+	my $feedonly=yesno($params{feedonly});
 	if (! exists $params{show} && ! $archive) {
 		$params{show}=10;
 	}
+	if (! exists $params{feedshow} && exists $params{show}) {
+		$params{feedshow}=$params{show};
+	}
 	my $desc;
 	if (exists $params{description}) {
 		$desc = $params{description} 
-	} else {
+	}
+	else {
 		$desc = $config{wikiname};
 	}
 	my $actions=yesno($params{actions});
@@ -114,6 +162,17 @@ sub preprocess_inline (@) { #{{{
 		@list=@list[$params{skip} .. scalar @list - 1];
 	}
 	
+	my @feedlist;
+	if ($feeds) {
+		if (exists $params{feedshow} &&
+		    $params{feedshow} && @list > $params{feedshow}) {
+			@feedlist=@list[0..$params{feedshow} - 1];
+		}
+		else {
+			@feedlist=@list;
+		}
+	}
+	
 	if ($params{show} && @list > $params{show}) {
 		@list=@list[0..$params{show} - 1];
 	}
@@ -122,13 +181,33 @@ sub preprocess_inline (@) { #{{{
 	# Explicitly add all currently displayed pages as dependencies, so
 	# that if they are removed or otherwise changed, the inline will be
 	# sure to be updated.
-	add_depends($params{page}, join(" or ", @list));
+	add_depends($params{page}, join(" or ", $#list >= $#feedlist ? @list : @feedlist));
 
-	my $rssurl=basename(rsspage($params{page}));
-	my $atomurl=basename(atompage($params{page}));
+	my $feednum="";
+
+	my $feedid=join("\0", map { $_."\0".$params{$_} } sort keys %params);
+	if (exists $knownfeeds{$feedid}) {
+		$feednum=$knownfeeds{$feedid};
+	}
+	else {
+		if (exists $page_numfeeds{$params{destpage}}) {
+			if ($feeds) {
+				$feednum=$knownfeeds{$feedid}=++$page_numfeeds{$params{destpage}};
+			}
+		}
+		else {
+			$feednum=$knownfeeds{$feedid}="";
+			if ($feeds) {
+				$page_numfeeds{$params{destpage}}=1;
+			}
+		}
+	}
+
+	my $rssurl=basename(rsspage($params{destpage}).$feednum) if $feeds && $rss;
+	my $atomurl=basename(atompage($params{destpage}).$feednum) if $feeds && $atom;
 	my $ret="";
 
-	if ($config{cgiurl} && (exists $params{rootpage} ||
+	if ($config{cgiurl} && ! $params{preview} && (exists $params{rootpage} ||
 			(exists $params{postform} && yesno($params{postform})))) {
 		# Add a blog post form, with feed buttons.
 		my $formtemplate=template("blogpost.tmpl", blind_cache => 1);
@@ -147,7 +226,7 @@ sub preprocess_inline (@) { #{{{
 		}
 		$ret.=$formtemplate->output;
 	}
-	elsif ($feeds) {
+	elsif ($feeds && !$params{preview}) {
 		# Add feed buttons.
 		my $linktemplate=template("feedlink.tmpl", blind_cache => 1);
 		$linktemplate->param(rssurl => $rssurl) if $rss;
@@ -155,97 +234,105 @@ sub preprocess_inline (@) { #{{{
 		$ret.=$linktemplate->output;
 	}
 	
-	require HTML::Template;
-	my @params=IkiWiki::template_params($params{template}.".tmpl", blind_cache => 1);
-	if (! @params) {
-		return sprintf(gettext("nonexistant template %s"), $params{template});
-	}
-	my $template=HTML::Template->new(@params) unless $raw;
+	if (! $feedonly) {
+		require HTML::Template;
+		my @params=IkiWiki::template_params($params{template}.".tmpl", blind_cache => 1);
+		if (! @params) {
+			return sprintf(gettext("nonexistant template %s"), $params{template});
+		}
+		my $template=HTML::Template->new(@params) unless $raw;
 	
-	foreach my $page (@list) {
-		my $file = $pagesources{$page};
-		my $type = pagetype($file);
-		if (! $raw || ($raw && ! defined $type)) {
-			unless ($archive && $quick) {
-				# Get the content before populating the
-				# template, since getting the content uses
-				# the same template if inlines are nested.
-				my $content=get_inline_content($page, $params{destpage});
-				$template->param(content => $content);
-			}
-			$template->param(pageurl => urlto(bestlink($params{page}, $page), $params{destpage}));
-			$template->param(title => pagetitle(basename($page)));
-			$template->param(ctime => displaytime($pagectime{$page}));
-
-			if ($actions) {
-				my $file = $pagesources{$page};
-				my $type = pagetype($file);
-				if ($config{discussion}) {
-					my $discussionlink=gettext("discussion");
-					if ($page !~ /.*\/\Q$discussionlink\E$/ &&
-					    (length $config{cgiurl} ||
-					     exists $links{$page."/".$discussionlink})) {
+		foreach my $page (@list) {
+			my $file = $pagesources{$page};
+			my $type = pagetype($file);
+			if (! $raw || ($raw && ! defined $type)) {
+				unless ($archive && $quick) {
+					# Get the content before populating the
+					# template, since getting the content uses
+					# the same template if inlines are nested.
+					my $content=get_inline_content($page, $params{destpage});
+					$template->param(content => $content);
+				}
+				$template->param(pageurl => urlto(bestlink($params{page}, $page), $params{destpage}));
+				$template->param(title => pagetitle(basename($page)));
+				$template->param(ctime => displaytime($pagectime{$page}, $params{timeformat}));
+				$template->param(first => 1) if $page eq $list[0];
+				$template->param(last => 1) if $page eq $list[$#list];
+	
+				if ($actions) {
+					my $file = $pagesources{$page};
+					my $type = pagetype($file);
+					if ($config{discussion}) {
+						my $discussionlink=gettext("discussion");
+						if ($page !~ /.*\/\Q$discussionlink\E$/ &&
+						    (length $config{cgiurl} ||
+						     exists $links{$page."/".$discussionlink})) {
+							$template->param(have_actions => 1);
+							$template->param(discussionlink =>
+								htmllink($page,
+									$params{destpage},
+									gettext("Discussion"),
+									noimageinline => 1,
+									forcesubpage => 1));
+						}
+					}
+					if (length $config{cgiurl} && defined $type) {
 						$template->param(have_actions => 1);
-						$template->param(discussionlink =>
-							htmllink($page,
-								$params{page},
-								gettext("Discussion"),
-								noimageinline => 1,
-								forcesubpage => 1));
+						$template->param(editurl => cgiurl(do => "edit", page => $page));
 					}
 				}
-				if (length $config{cgiurl} && defined $type) {
-					$template->param(have_actions => 1);
-					$template->param(editurl => cgiurl(do => "edit", page => pagetitle($page, 1)));
-				}
+	
+				run_hooks(pagetemplate => sub {
+					shift->(page => $page, destpage => $params{destpage},
+						template => $template,);
+				});
+	
+				$ret.=$template->output;
+				$template->clear_params;
 			}
-
-			run_hooks(pagetemplate => sub {
-				shift->(page => $page, destpage => $params{page},
-					template => $template,);
-			});
-
-			$ret.=$template->output;
-			$template->clear_params;
-		}
-		else {
-			if (defined $type) {
-				$ret.="\n".
-				      linkify($page, $params{page},
-				      preprocess($page, $params{page},
-				      filter($page, $params{page},
-				      readfile(srcfile($file)))));
+			else {
+				if (defined $type) {
+					$ret.="\n".
+					      linkify($page, $params{destpage},
+					      preprocess($page, $params{destpage},
+					      filter($page, $params{destpage},
+					      readfile(srcfile($file)))));
+				}
 			}
 		}
 	}
 	
 	if ($feeds) {
-		if (exists $params{feedshow} && @list > $params{feedshow}) {
-			@list=@list[0..$params{feedshow} - 1];
-		}
 		if (exists $params{feedpages}) {
-			@list=grep { pagespec_match($_, $params{feedpages}, location => $params{page}) } @list;
+			@feedlist=grep { pagespec_match($_, $params{feedpages}, location => $params{page}) } @feedlist;
 		}
 	
 		if ($rss) {
-			my $rssp=rsspage($params{page});
-			will_render($params{page}, $rssp);
-			writefile($rssp, $config{destdir},
-				genfeed("rss", $rssurl, $desc, $params{page}, @list));
-			$toping{$params{page}}=1 unless $config{rebuild};
-			$feedlinks{$params{destpage}}=qq{<link rel="alternate" type="application/rss+xml" title="RSS" href="$rssurl" />};
+			my $rssp=rsspage($params{destpage}).$feednum;
+			will_render($params{destpage}, $rssp);
+			if (! $params{preview}) {
+				writefile($rssp, $config{destdir},
+					genfeed("rss",
+						$config{url}."/".rsspage($params{destpage}).$feednum, $desc, $params{destpage}, @feedlist));
+				$toping{$params{destpage}}=1 unless $config{rebuild};
+				$feedlinks{$params{destpage}}=qq{<link rel="alternate" type="application/rss+xml" title="RSS" href="$rssurl" />};
+			}
 		}
 		if ($atom) {
-			my $atomp=atompage($params{page});
-			will_render($params{page}, $atomp);
-			writefile($atomp, $config{destdir},
-				genfeed("atom", $atomurl, $desc, $params{page}, @list));
-			$toping{$params{page}}=1 unless $config{rebuild};
-			$feedlinks{$params{destpage}}=qq{<link rel="alternate" type="application/atom+xml" title="Atom" href="$atomurl" />};
+			my $atomp=atompage($params{destpage}).$feednum;
+			will_render($params{destpage}, $atomp);
+			if (! $params{preview}) {
+				writefile($atomp, $config{destdir},
+					genfeed("atom", $config{url}."/".atompage($params{destpage}).$feednum, $desc, $params{destpage}, @feedlist));
+				$toping{$params{destpage}}=1 unless $config{rebuild};
+				$feedlinks{$params{destpage}}=qq{<link rel="alternate" type="application/atom+xml" title="Atom" href="$atomurl" />};
+			}
 		}
 	}
 	
-	return $ret;
+	return $ret if $raw || $nested;
+	push @inline, $ret;
+	return "<div class=\"inline\" id=\"$#inline\"></div>\n\n";
 } #}}}
 
 sub pagetemplate_inline (@) { #{{{
@@ -264,11 +351,14 @@ sub get_inline_content ($$) { #{{{
 	my $file=$pagesources{$page};
 	my $type=pagetype($file);
 	if (defined $type) {
-		return htmlize($page, $type,
+		$nested++;
+		my $ret=htmlize($page, $destpage, $type,
 		       linkify($page, $destpage,
 		       preprocess($page, $destpage,
 		       filter($page, $destpage,
 		       readfile(srcfile($file))))));
+		$nested--;
+		return $ret;
 	}
 	else {
 		return "";
@@ -290,7 +380,7 @@ sub date_3339 ($) { #{{{
 
 	my $lc_time=POSIX::setlocale(&POSIX::LC_TIME);
 	POSIX::setlocale(&POSIX::LC_TIME, "C");
-	my $ret=POSIX::strftime("%Y-%m-%dT%H:%M:%SZ", localtime($time));
+	my $ret=POSIX::strftime("%Y-%m-%dT%H:%M:%SZ", gmtime($time));
 	POSIX::setlocale(&POSIX::LC_TIME, $lc_time);
 	return $ret;
 } #}}}
@@ -302,10 +392,19 @@ sub absolute_urls ($$) { #{{{
 
 	my $url=$baseurl;
 	$url=~s/[^\/]+$//;
-	
+
+        # what is the non path part of the url?
+        my $top_uri = URI->new($url);
+        $top_uri->path_query(""); # reset the path
+        my $urltop = $top_uri->as_string;
+
 	$content=~s/(<a(?:\s+(?:class|id)\s*="?\w+"?)?)\s+href=\s*"(#[^"]+)"/$1 href="$baseurl$2"/mig;
-	$content=~s/(<a(?:\s+(?:class|id)\s*="?\w+"?)?)\s+href=\s*"(?!\w+:\/\/)([^"]+)"/$1 href="$url$2"/mig;
-	$content=~s/(<img(?:\s+(?:class|id|width|height)\s*="?\w+"?)*)\s+src=\s*"(?!\w+:\/\/)([^"]+)"/$1 src="$url$2"/mig;
+        # relative to another wiki page
+	$content=~s/(<a(?:\s+(?:class|id)\s*="?\w+"?)?)\s+href=\s*"(?!\w+:)([^\/][^"]*)"/$1 href="$url$2"/mig;
+	$content=~s/(<img(?:\s+(?:class|id|width|height)\s*="?\w+"?)*)\s+src=\s*"(?!\w+:)([^\/][^"]*)"/$1 src="$url$2"/mig;
+        # relative to the top of the site
+	$content=~s/(<a(?:\s+(?:class|id)\s*="?\w+"?)?)\s+href=\s*"(?!\w+:)(\/[^"]*)"/$1 href="$urltop$2"/mig;
+	$content=~s/(<img(?:\s+(?:class|id|width|height)\s*="?\w+"?)*)\s+src=\s*"(?!\w+:)(\/[^"]*)"/$1 src="$urltop$2"/mig;
 	return $content;
 } #}}}
 
@@ -337,8 +436,10 @@ sub genfeed ($$$$@) { #{{{
 			title => pagetitle(basename($p)),
 			url => $u,
 			permalink => $u,
-			date_822 => date_822($pagectime{$p}),
-			date_3339 => date_3339($pagectime{$p}),
+			cdate_822 => date_822($pagectime{$p}),
+			mdate_822 => date_822($pagemtime{$p}),
+			cdate_3339 => date_3339($pagectime{$p}),
+			mdate_3339 => date_3339($pagemtime{$p}),
 		);
 
 		if ($itemtemplate->query(name => "enclosure")) {
@@ -348,7 +449,7 @@ sub genfeed ($$$$@) { #{{{
 				$itemtemplate->param(content => $pcontent);
 			}
 			else {
-				my ($a, $b, $c, $d, $e, $f, $g, $size) = stat(srcfile($file));
+				my $size=(srcfile_stat($file))[8];
 				my $mime="unknown";
 				eval q{use File::MimeInfo};
 				if (! $@) {
@@ -373,7 +474,7 @@ sub genfeed ($$$$@) { #{{{
 		$content.=$itemtemplate->output;
 		$itemtemplate->clear_params;
 
-		$lasttime = $pagectime{$p} if $pagectime{$p} > $lasttime;
+		$lasttime = $pagemtime{$p} if $pagemtime{$p} > $lasttime;
 	}
 
 	my $template=template($feedtype."page.tmpl", blind_cache => 1);
@@ -408,7 +509,7 @@ sub pingurl (@) { #{{{
 	defined(my $pid = fork) or error("Can't fork: $!");
 	return if $pid;
 	chdir '/';
-	setsid() or error("Can't start a new session: $!");
+	POSIX::setsid() or error("Can't start a new session: $!");
 	open STDIN, '/dev/null';
 	open STDOUT, '>/dev/null';
 	open STDERR, '>&STDOUT' or error("Can't dup stdout: $!");
