@@ -55,8 +55,8 @@ sub parentlinks ($) { #{{{
 	my $path="";
 	my $title=$config{wikiname};
 	
-	return if $page eq 'index'; # toplevel
 	foreach my $dir (split("/", $page)) {
+		next if $dir eq 'index';
 		push @ret, { url => urlto($path, $page), page => $title };
 		$path.="/".$dir;
 		$title=pagetitle($dir);
@@ -64,23 +64,27 @@ sub parentlinks ($) { #{{{
 	return @ret;
 } #}}}
 
-sub genpage ($$$) { #{{{
+sub genpage ($$) { #{{{
 	my $page=shift;
 	my $content=shift;
-	my $mtime=shift;
 
-	my $template=template("page.tmpl", blind_cache => 1);
+	my $templatefile;
+	run_hooks(templatefile => sub {
+		return if defined $templatefile;
+		my $file=shift->(page => $page);
+		if (defined $file && defined template_file($file)) {
+			$templatefile=$file;
+		}
+	});
+	my $template=template(defined $templatefile ? $templatefile : 'page.tmpl', blind_cache => 1);
 	my $actions=0;
 
 	if (length $config{cgiurl}) {
-		$template->param(editurl => cgiurl(do => "edit", page => pagetitle($page, 1)));
+		$template->param(editurl => cgiurl(do => "edit", page => $page));
 		$template->param(prefsurl => cgiurl(do => "prefs"));
-		if ($config{rcs}) {
-			$template->param(recentchangesurl => cgiurl(do => "recentchanges"));
-		}
 		$actions++;
 	}
-
+		
 	if (length $config{historyurl}) {
 		my $u=$config{historyurl};
 		$u=~s/\[\[file\]\]/$pagesources{$page}/g;
@@ -121,7 +125,8 @@ sub genpage ($$$) { #{{{
 		content => $content,
 		backlinks => $backlinks,
 		more_backlinks => $more_backlinks,
-		mtime => displaytime($mtime),
+		mtime => displaytime($pagemtime{$page}),
+		ctime => displaytime($pagectime{$page}),
 		baseurl => baseurl($page),
 	);
 
@@ -141,12 +146,6 @@ sub genpage ($$$) { #{{{
 	return $content;
 } #}}}
 
-sub mtime ($) { #{{{
-	my $file=shift;
-	
-	return (stat($file))[9];
-} #}}}
-
 sub scan ($) { #{{{
 	my $file=shift;
 
@@ -157,21 +156,22 @@ sub scan ($) { #{{{
 		my $page=pagename($file);
 		will_render($page, htmlpage($page), 1);
 
-		# Always needs to be done, since filters might add links
-		# to the content.
-		$content=filter($page, $page, $content);
-
-		my @links;
-		while ($content =~ /(?<!\\)$config{wiki_link_regexp}/g) {
-			push @links, linkpage($2);
-		}
 		if ($config{discussion}) {
 			# Discussion links are a special case since they're
 			# not in the text of the page, but on its template.
-			push @links, $page."/".gettext("discussion");
+			$links{$page}=[ $page."/".gettext("discussion") ];
 		}
-		$links{$page}=\@links;
-		
+		else {
+			$links{$page}=[];
+		}
+
+		run_hooks(scan => sub {
+			shift->(
+				page => $page,
+				content => $content,
+			);
+		});
+
 		# Preprocess in scan-only mode.
 		preprocess($page, $page, $content, 1);
 	}
@@ -180,6 +180,30 @@ sub scan ($) { #{{{
 	}
 } #}}}
 
+sub fast_file_copy (@) { #{{{
+	my $srcfile=shift;
+	my $destfile=shift;
+	my $srcfd=shift;
+	my $destfd=shift;
+	my $cleanup=shift;
+
+	my $blksize = 16384;
+	my ($len, $buf, $written);
+	while ($len = sysread $srcfd, $buf, $blksize) {
+		if (! defined $len) {
+			next if $! =~ /^Interrupted/;
+			error("failed to read $srcfile: $!", $cleanup);
+		}
+		my $offset = 0;
+		while ($len) {
+			defined($written = syswrite $destfd, $buf, $len, $offset)
+				or error("failed to write $destfile: $!", $cleanup);
+			$len -= $written;
+			$offset += $written;
+		}
+	}
+}
+
 sub render ($) { #{{{
 	my $file=shift;
 	
@@ -189,39 +213,33 @@ sub render ($) { #{{{
 		my $page=pagename($file);
 		delete $depends{$page};
 		will_render($page, htmlpage($page), 1);
+		return if $type=~/^_/;
 		
-		my $content=htmlize($page, $type,
+		my $content=htmlize($page, $page, $type,
 			linkify($page, $page,
 			preprocess($page, $page,
 			filter($page, $page,
 			readfile($srcfile)))));
 		
-		writefile(htmlpage($page), $config{destdir},
-			genpage($page, $content, mtime($srcfile)));
+		my $output=htmlpage($page);
+		writefile($output, $config{destdir}, genpage($page, $content));
 	}
 	else {
-		my $srcfd=readfile($srcfile, 1, 1);
 		delete $depends{$file};
 		will_render($file, $file, 1);
-		writefile($file, $config{destdir}, undef, 1, sub {
-			my $destfd=shift;
-			my $cleanup=shift;
-
-			my $blksize = 16384;
-			my ($len, $buf, $written);
-			while ($len = sysread $srcfd, $buf, $blksize) {
-				if (! defined $len) {
-					next if $! =~ /^Interrupted/;
-					error("failed to read $srcfile: $!", $cleanup);
-				}
-				my $offset = 0;
-				while ($len) {
-					defined($written = syswrite $destfd, $buf, $len, $offset)
-						or error("failed to write $file: $!", $cleanup);
-					$len -= $written;
-					$offset += $written;
-				}
+		
+		if ($config{hardlink}) {
+			prep_writefile($file, $config{destdir});
+			unlink($config{destdir}."/".$file);
+			if (link($srcfile, $config{destdir}."/".$file)) {
+				return;
 			}
+			# if hardlink fails, fall back to copying
+		}
+		
+		my $srcfd=readfile($srcfile, 1, 1);
+		writefile($file, $config{destdir}, undef, 1, sub {
+			fast_file_copy($srcfile, $file, $srcfd, @_);
 		});
 	}
 } #}}}
@@ -237,6 +255,19 @@ sub prune ($) { #{{{
 } #}}}
 
 sub refresh () { #{{{
+	# security check, avoid following symlinks in the srcdir path
+	my $test=$config{srcdir};
+	while (length $test) {
+		if (-l $test) {
+			error("symlink found in srcdir path ($test)");
+		}
+		unless ($test=~s/\/+$//) {
+			$test=dirname($test);
+		}
+	}
+	
+	run_hooks(refresh => sub { shift->() });
+
 	# find existing pages
 	my %exists;
 	my @files;
@@ -249,7 +280,7 @@ sub refresh () { #{{{
 			if (file_pruned($_, $config{srcdir})) {
 				$File::Find::prune=1;
 			}
-			elsif (! -d $_ && ! -l $_) {
+			elsif (! -l $_ && ! -d _) {
 				my ($f)=/$config{wiki_file_regexp}/; # untaint
 				if (! defined $f) {
 					warn(sprintf(gettext("skipping bad filename %s"), $_)."\n");
@@ -262,64 +293,85 @@ sub refresh () { #{{{
 			}
 		},
 	}, $config{srcdir});
-	find({
-		no_chdir => 1,
-		wanted => sub {
-			$_=decode_utf8($_);
-			if (file_pruned($_, $config{underlaydir})) {
-				$File::Find::prune=1;
-			}
-			elsif (! -d $_ && ! -l $_) {
-				my ($f)=/$config{wiki_file_regexp}/; # untaint
-				if (! defined $f) {
-					warn(sprintf(gettext("skipping bad filename %s"), $_)."\n");
+	foreach my $dir (@{$config{underlaydirs}}, $config{underlaydir}) {
+		find({
+			no_chdir => 1,
+			wanted => sub {
+				$_=decode_utf8($_);
+				if (file_pruned($_, $dir)) {
+					$File::Find::prune=1;
 				}
-				else {
-					# Don't add pages that are in the
-					# srcdir.
-					$f=~s/^\Q$config{underlaydir}\E\/?//;
-					if (! -e "$config{srcdir}/$f" && 
-					    ! -l "$config{srcdir}/$f") {
-					    	my $page=pagename($f);
-						if (! $exists{$page}) {
-							push @files, $f;
-							$exists{$page}=1;
+				elsif (! -l $_ && ! -d _) {
+					my ($f)=/$config{wiki_file_regexp}/; # untaint
+					if (! defined $f) {
+						warn(sprintf(gettext("skipping bad filename %s"), $_)."\n");
+					}
+					else {
+						$f=~s/^\Q$dir\E\/?//;
+						# avoid underlaydir
+						# override attacks; see
+						# security.mdwn
+						if (! -l "$config{srcdir}/$f" && 
+						    ! -e _) {
+						    	my $page=pagename($f);
+							if (! $exists{$page}) {
+								push @files, $f;
+								$exists{$page}=1;
+							}
 						}
 					}
 				}
-			}
-		},
-	}, $config{underlaydir});
-
-	my %rendered;
+			},
+		}, $dir);
+	};
 
+	my (%rendered, @add, @del, @internal);
 	# check for added or removed pages
-	my @add;
 	foreach my $file (@files) {
 		my $page=pagename($file);
 		$pagesources{$page}=$file;
 		if (! $pagemtime{$page}) {
-			push @add, $file;
-			$pagecase{lc $page}=$page;
-			if ($config{getctime} && -e "$config{srcdir}/$file") {
-				$pagectime{$page}=rcs_getctime("$config{srcdir}/$file");
+			if (isinternal($page)) {
+				push @internal, $file;
 			}
-			elsif (! exists $pagectime{$page}) {
-				$pagectime{$page}=mtime(srcfile($file));
+			else {
+				push @add, $file;
+				if ($config{getctime} && -e "$config{srcdir}/$file") {
+					eval {
+						my $time=rcs_getctime("$config{srcdir}/$file");
+						$pagectime{$page}=$time;
+					};
+					if ($@) {
+						print STDERR $@;
+					}
+				}
+			}
+			$pagecase{lc $page}=$page;
+			if (! exists $pagectime{$page}) {
+				$pagectime{$page}=(srcfile_stat($file))[10];
 			}
 		}
 	}
-	my @del;
 	foreach my $page (keys %pagemtime) {
 		if (! $exists{$page}) {
-			debug(sprintf(gettext("removing old page %s"), $page));
-			push @del, $pagesources{$page};
+			if (isinternal($page)) {
+				push @internal, $pagesources{$page};
+			}
+			else {
+				debug(sprintf(gettext("removing old page %s"), $page));
+				push @del, $pagesources{$page};
+			}
 			$links{$page}=[];
 			$renderedfiles{$page}=[];
 			$pagemtime{$page}=0;
 			prune($config{destdir}."/".$_)
 				foreach @{$oldrenderedfiles{$page}};
 			delete $pagesources{$page};
+			foreach (keys %destsources) {
+				if ($destsources{$_} eq $page) {
+					delete $destsources{$_};
+				}
+			}
 		}
 	}
 
@@ -327,18 +379,24 @@ sub refresh () { #{{{
 	my @needsbuild;
 	foreach my $file (@files) {
 		my $page=pagename($file);
-		
-		my $mtime=mtime(srcfile($file));
+		my ($srcfile, @stat)=srcfile_stat($file);
 		if (! exists $pagemtime{$page} ||
-		    $mtime > $pagemtime{$page} ||
+		    $stat[9] > $pagemtime{$page} ||
 	    	    $forcerebuild{$page}) {
-			$pagemtime{$page}=$mtime;
-			push @needsbuild, $file;
+			$pagemtime{$page}=$stat[9];
+			if (isinternal($page)) {
+				push @internal, $file;
+				# Preprocess internal page in scan-only mode.
+				preprocess($page, $page, readfile($srcfile), 1);
+			}
+			else {
+				push @needsbuild, $file;
+			}
 		}
 	}
 	run_hooks(needsbuild => sub { shift->(\@needsbuild) });
 
-	# scan and rendder files
+	# scan and render files
 	foreach my $file (@needsbuild) {
 		debug(sprintf(gettext("scanning %s"), $file));
 		scan($file);
@@ -349,6 +407,15 @@ sub refresh () { #{{{
 		render($file);
 		$rendered{$file}=1;
 	}
+	foreach my $file (@internal) {
+		# internal pages are not rendered
+		my $page=pagename($file);
+		delete $depends{$page};
+		foreach my $old (@{$renderedfiles{$page}}) {
+			delete $destsources{$old};
+		}
+		$renderedfiles{$page}=[];
+	}
 	
 	# rebuild pages that link to added or removed pages
 	if (@add || @del) {
@@ -364,13 +431,17 @@ sub refresh () { #{{{
 		}
 	}
 
-	if (%rendered || @del) {
+	if (%rendered || @del || @internal) {
+		my @changed=(keys %rendered, @del);
+
 		# rebuild dependant pages
 		foreach my $f (@files) {
 			next if $rendered{$f};
 			my $p=pagename($f);
 			if (exists $depends{$p}) {
-				foreach my $file (keys %rendered, @del) {
+				# only consider internal files
+				# if the page explicitly depends on such files
+				foreach my $file (@changed, $depends{$p}=~/internal\(/ ? @internal : ()) {
 					next if $f eq $file;
 					my $page=pagename($file);
 					if (pagespec_match($page, $depends{$p}, location => $p)) {
@@ -386,7 +457,7 @@ sub refresh () { #{{{
 		# handle backlinks; if a page has added/removed links,
 		# update the pages it links to
 		my %linkchanged;
-		foreach my $file (keys %rendered, @del) {
+		foreach my $file (@changed) {
 			my $page=pagename($file);
 			
 			if (exists $links{$page}) {
@@ -408,6 +479,7 @@ sub refresh () { #{{{
 				}
 			}
 		}
+
 		foreach my $link (keys %linkchanged) {
 		    	my $linkfile=$pagesources{$link};
 			if (defined $linkfile) {
@@ -439,8 +511,6 @@ sub refresh () { #{{{
 } #}}}
 
 sub commandline_render () { #{{{
-	loadplugins();
-	checkconfig();
 	lockwiki();
 	loadindex();
 	unlockwiki();
@@ -457,9 +527,10 @@ sub commandline_render () { #{{{
 	$content=filter($page, $page, $content);
 	$content=preprocess($page, $page, $content);
 	$content=linkify($page, $page, $content);
-	$content=htmlize($page, $type, $content);
+	$content=htmlize($page, $page, $type, $content);
+	$pagemtime{$page}=(stat($srcfile))[9];
 
-	print genpage($page, $content, mtime($srcfile));
+	print genpage($page, $content);
 	exit 0;
 } #}}}
 
